from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from logging import getLogger
from os.path import dirname, join
from pathlib import Path
from re import IGNORECASE
from typing import Any, List, Optional, Tuple, Union

from lark import Lark, ParseTree, Token, Transformer, Tree, v_args
from lark.exceptions import (
    UnexpectedCharacters,
    UnexpectedEOF,
    UnexpectedInput,
    UnexpectedToken,
    VisitError,
)
from lark.tree import Meta
from pydantic import ValidationError

from trilogy.constants import (
    CONFIG,
    DEFAULT_NAMESPACE,
    NULL_VALUE,
    MagicConstants,
    Parsing,
)
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    ConceptSource,
    DatePart,
    Derivation,
    FunctionType,
    Granularity,
    IOType,
    Modifier,
    Ordering,
    Purpose,
    ShowCategory,
    ValidationScope,
    WindowOrder,
    WindowType,
)
from trilogy.core.exceptions import InvalidSyntaxException, UndefinedConceptException
from trilogy.core.functions import (
    CurrentDate,
    FunctionFactory,
)
from trilogy.core.internal import ALL_ROWS_CONCEPT, INTERNAL_NAMESPACE
from trilogy.core.models.author import (
    AggregateWrapper,
    AlignClause,
    AlignItem,
    ArgBinding,
    CaseElse,
    CaseWhen,
    Comment,
    Comparison,
    Concept,
    ConceptRef,
    Conditional,
    CustomFunctionFactory,
    CustomType,
    DeriveClause,
    DeriveItem,
    Expr,
    FilterItem,
    Function,
    FunctionCallWrapper,
    Grain,
    HavingClause,
    Metadata,
    MultiSelectLineage,
    OrderBy,
    OrderItem,
    Parenthetical,
    RowsetItem,
    SubselectComparison,
    UndefinedConceptFull,
    WhereClause,
    Window,
    WindowItem,
    WindowItemOrder,
    WindowItemOver,
)
from trilogy.core.models.core import (
    ArrayType,
    DataType,
    DataTyped,
    ListWrapper,
    MapType,
    MapWrapper,
    NumericType,
    StructComponent,
    StructType,
    TraitDataType,
    TupleWrapper,
    arg_to_datatype,
    dict_to_map_wrapper,
    is_compatible_datatype,
    list_to_wrapper,
    tuple_to_wrapper,
)
from trilogy.core.models.datasource import (
    Address,
    ColumnAssignment,
    Datasource,
    Query,
    RawColumnExpr,
)
from trilogy.core.models.environment import (
    DictImportResolver,
    Environment,
    FileSystemImportResolver,
    Import,
)
from trilogy.core.statements.author import (
    ConceptDeclarationStatement,
    ConceptDerivationStatement,
    ConceptTransform,
    CopyStatement,
    FunctionDeclaration,
    ImportStatement,
    Limit,
    MergeStatementV2,
    MultiSelectStatement,
    PersistStatement,
    RawSQLStatement,
    RowsetDerivationStatement,
    SelectItem,
    SelectStatement,
    ShowStatement,
    TypeDeclaration,
    ValidateStatement,
)
from trilogy.parsing.common import (
    align_item_to_concept,
    arbitrary_to_concept,
    constant_to_concept,
    derive_item_to_concept,
    process_function_args,
    rowset_to_concepts,
)
from trilogy.parsing.exceptions import NameShadowError, ParseError

perf_logger = getLogger("trilogy.parse.performance")


class ParsePass(Enum):
    INITIAL = 1
    VALIDATION = 2


CONSTANT_TYPES = (int, float, str, bool, ListWrapper, TupleWrapper, MapWrapper)

SELF_LABEL = "root"

MAX_PARSE_DEPTH = 10

STDLIB_ROOT = Path(__file__).parent.parent


@dataclass
class WholeGrainWrapper:
    where: WhereClause


@dataclass
class FunctionBindingType:
    type: DataType | TraitDataType | None = None


@dataclass
class DropOn:
    functions: List[FunctionType]


@dataclass
class AddOn:
    functions: List[FunctionType]


with open(join(dirname(__file__), "trilogy.lark"), "r") as f:
    PARSER = Lark(
        f.read(),
        start="start",
        propagate_positions=True,
        g_regex_flags=IGNORECASE,
        parser="lalr",
        cache=True,
    )


def parse_concept_reference(
    name: str, environment: Environment, purpose: Optional[Purpose] = None
) -> Tuple[str, str, str, str | None]:
    parent = None
    if "." in name:
        if purpose == Purpose.PROPERTY:
            parent, name = name.rsplit(".", 1)
            namespace = environment.concepts[parent].namespace or DEFAULT_NAMESPACE
            lookup = f"{namespace}.{name}"
        else:
            namespace, name = name.rsplit(".", 1)
            lookup = f"{namespace}.{name}"
    else:
        namespace = environment.namespace or DEFAULT_NAMESPACE
        lookup = name
    return lookup, namespace, name, parent


def expr_to_boolean(
    root,
    function_factory: FunctionFactory,
) -> Union[Comparison, SubselectComparison, Conditional]:
    if not isinstance(root, (Comparison, SubselectComparison, Conditional)):
        if arg_to_datatype(root) == DataType.BOOL:
            root = Comparison(left=root, right=True, operator=ComparisonOperator.EQ)
        elif arg_to_datatype(root) == DataType.INTEGER:
            root = Comparison(
                left=function_factory.create_function(
                    [root],
                    FunctionType.BOOL,
                ),
                right=True,
                operator=ComparisonOperator.EQ,
            )
        else:
            root = Comparison(
                left=root, right=NULL_VALUE, operator=ComparisonOperator.IS_NOT
            )

    return root


def unwrap_transformation(
    input: Expr,
    environment: Environment,
) -> (
    Function
    | FilterItem
    | WindowItem
    | AggregateWrapper
    | FunctionCallWrapper
    | Parenthetical
):
    if isinstance(input, Function):
        return input
    elif isinstance(input, AggregateWrapper):
        return input
    elif isinstance(input, ConceptRef):
        concept = environment.concepts[input.address]
        return Function(
            operator=FunctionType.ALIAS,
            output_datatype=concept.datatype,
            output_purpose=concept.purpose,
            arguments=[input],
        )
    elif isinstance(input, FilterItem):
        return input
    elif isinstance(input, WindowItem):
        return input
    elif isinstance(input, FunctionCallWrapper):
        return input
    elif isinstance(input, Parenthetical):
        return input
    else:
        return Function.model_construct(
            operator=FunctionType.CONSTANT,
            output_datatype=arg_to_datatype(input),
            output_purpose=Purpose.CONSTANT,
            arguments=[input],
        )


def rehydrate_lineage(
    lineage: Any, environment: Environment, function_factory: FunctionFactory
) -> Any:
    """Fix datatype propagation. This is a hack to fix the fact that we don't know the datatypes of functions until we've parsed all concepts"""
    if isinstance(lineage, Function):
        rehydrated = [
            rehydrate_lineage(x, environment, function_factory)
            for x in lineage.arguments
        ]
        return function_factory.create_function(
            rehydrated,
            operator=lineage.operator,
        )
    elif isinstance(lineage, Parenthetical):
        lineage.content = rehydrate_lineage(
            lineage.content, environment, function_factory
        )
        return lineage
    elif isinstance(lineage, WindowItem):
        # this is temporarily guaranteed until we do some upstream work
        assert isinstance(lineage.content, ConceptRef)
        lineage.content.datatype = environment.concepts[
            lineage.content.address
        ].datatype
        return lineage
    elif isinstance(lineage, AggregateWrapper):
        lineage.function = rehydrate_lineage(
            lineage.function, environment, function_factory
        )
        return lineage
    elif isinstance(lineage, RowsetItem):
        lineage.content.datatype = environment.concepts[
            lineage.content.address
        ].datatype
        return lineage
    else:
        return lineage


def rehydrate_concept_lineage(
    concept: Concept, environment: Environment, function_factory: FunctionFactory
) -> Concept:
    concept.lineage = rehydrate_lineage(concept.lineage, environment, function_factory)
    if isinstance(concept.lineage, DataTyped):
        concept.datatype = concept.lineage.output_datatype
    return concept


class ParseToObjects(Transformer):
    def __init__(
        self,
        environment: Environment,
        parse_address: str | None = None,
        token_address: Path | str | None = None,
        parsed: dict[str, "ParseToObjects"] | None = None,
        tokens: dict[Path | str, ParseTree] | None = None,
        text_lookup: dict[Path | str, str] | None = None,
        environment_lookup: dict[str, Environment] | None = None,
        import_keys: list[str] | None = None,
        parse_config: Parsing | None = None,
    ):
        Transformer.__init__(self, True)
        self.environment: Environment = environment
        self.parse_address: str = parse_address or SELF_LABEL
        self.token_address: Path | str = token_address or SELF_LABEL
        self.parsed: dict[str, ParseToObjects] = parsed if parsed is not None else {}
        self.tokens: dict[Path | str, ParseTree] = tokens if tokens is not None else {}
        self.environments: dict[str, Environment] = environment_lookup or {}
        self.text_lookup: dict[Path | str, str] = (
            text_lookup if text_lookup is not None else {}
        )
        # we do a second pass to pick up circular dependencies
        # after initial parsing
        self.parse_pass = ParsePass.INITIAL
        self.function_factory = FunctionFactory(self.environment)
        self.import_keys: list[str] = import_keys or ["root"]
        self.parse_config: Parsing = parse_config or CONFIG.parsing

    def set_text(self, text: str):
        self.text_lookup[self.token_address] = text

    def transform(self, tree: Tree):
        results = super().transform(tree)
        self.tokens[self.token_address] = tree
        return results

    def prepare_parse(self):
        self.parse_pass = ParsePass.INITIAL
        self.environment.concepts.fail_on_missing = False
        for _, v in self.parsed.items():
            v.prepare_parse()

    def run_second_parse_pass(self, force: bool = False):
        if self.token_address not in self.tokens:
            return []
        self.parse_pass = ParsePass.VALIDATION
        for _, v in list(self.parsed.items()):
            if v.parse_pass == ParsePass.VALIDATION:
                continue
            v.run_second_parse_pass()
        reparsed = self.transform(self.tokens[self.token_address])
        self.environment.concepts.undefined = {}
        passed = False
        passes = 0
        # output datatypes for functions may have been wrong
        # as they were derived from not fully understood upstream types
        # so loop through to recreate function lineage until all datatypes are known

        while not passed:
            new_passed = True
            for x, y in self.environment.concepts.items():
                if y.datatype == DataType.UNKNOWN and y.lineage:
                    self.environment.concepts[x] = rehydrate_concept_lineage(
                        y, self.environment, self.function_factory
                    )
                    new_passed = False
            passes += 1
            if passes > MAX_PARSE_DEPTH:
                break
            passed = new_passed

        return reparsed

    def start(self, args):
        return args

    def LINE_SEPARATOR(self, args):
        return MagicConstants.LINE_SEPARATOR

    def block(self, args):
        output = args[0]
        if isinstance(output, ConceptDeclarationStatement):
            if len(args) > 1 and args[1] != MagicConstants.LINE_SEPARATOR:
                comments = [x for x in args[1:] if isinstance(x, Comment)]
                merged = "\n".join([x.text.split("#")[1].rstrip() for x in comments])
                output.concept.metadata.description = merged
        # this is a bad plan for now;
        # because a comment after an import statement is very common
        # and it's not intuitive that it modifies the import description
        # if isinstance(output, ImportStatement):
        #     if len(args) > 1 and isinstance(args[1], Comment):
        #         comment = args[1].text.split("#")[1].strip()
        #         namespace = output.alias
        #         for _, v in self.environment.concepts.items():
        #             if v.namespace == namespace:
        #                 if v.metadata.description:
        #                     v.metadata.description = (
        #                         f"{comment}: {v.metadata.description}"
        #                     )
        #                 else:
        #                     v.metadata.description = comment

        return args[0]

    def metadata(self, args):
        pairs = {key: val for key, val in zip(args[::2], args[1::2])}
        return Metadata(**pairs)

    def IDENTIFIER(self, args) -> str:
        return args.value

    def ORDER_IDENTIFIER(self, args) -> ConceptRef:
        return self.environment.concepts[args.value.strip()].reference

    def WILDCARD_IDENTIFIER(self, args) -> str:
        return args.value

    def QUOTED_IDENTIFIER(self, args) -> str:
        return args.value[1:-1]

    @v_args(meta=True)
    def concept_lit(self, meta: Meta, args) -> ConceptRef:
        address = args[0]
        if "." not in address and self.environment.namespace == DEFAULT_NAMESPACE:
            address = f"{DEFAULT_NAMESPACE}.{address}"
        mapping = self.environment.concepts[address]
        datatype = mapping.output_datatype
        return ConceptRef(
            # this is load-bearing to handle pseudonyms
            address=mapping.address,
            metadata=Metadata(line_number=meta.line),
            datatype=datatype,
        )

    def ADDRESS(self, args) -> Address:
        return Address(location=args.value, quoted=False)

    def QUOTED_ADDRESS(self, args) -> Address:
        return Address(location=args.value[1:-1], quoted=True)

    def STRING_CHARS(self, args) -> str:
        return args.value

    def SINGLE_STRING_CHARS(self, args) -> str:
        return args.value

    def DOUBLE_STRING_CHARS(self, args) -> str:
        return args.value

    def MINUS(self, args) -> str:
        return "-"

    @v_args(meta=True)
    def struct_component(self, meta: Meta, args) -> StructComponent:
        modifiers = []
        for arg in args:
            if isinstance(arg, Modifier):
                modifiers.append(arg)
        return StructComponent(name=args[0], type=args[1], modifiers=modifiers)

    @v_args(meta=True)
    def struct_type(self, meta: Meta, args) -> StructType:
        final: list[
            DataType
            | MapType
            | ArrayType
            | NumericType
            | StructType
            | StructComponent
            | Concept
        ] = []
        for arg in args:
            if isinstance(arg, StructComponent):
                final.append(arg)
            else:
                new = self.environment.concepts.__getitem__(  # type: ignore
                    key=arg, line_no=meta.line
                )
                final.append(new)

        return StructType(
            fields=final,
            fields_map={
                x.name: x for x in final if isinstance(x, (Concept, StructComponent))
            },
        )

    def list_type(self, args) -> ArrayType:
        content = args[0]
        if isinstance(content, str):
            content = self.environment.concepts[content]
        return ArrayType(type=content)

    def numeric_type(self, args) -> NumericType:
        return NumericType(precision=args[0], scale=args[1])

    def map_type(self, args) -> MapType:
        key = args[0]
        value = args[1]
        if isinstance(key, str):
            key = self.environment.concepts[key]
        elif isinstance(value, str):
            value = self.environment.concepts[value]
        return MapType(key_type=key, value_type=value)

    @v_args(meta=True)
    def data_type(
        self, meta: Meta, args
    ) -> DataType | TraitDataType | ArrayType | StructType | MapType | NumericType:
        resolved = args[0]
        traits = args[2:]
        base: DataType | TraitDataType | ArrayType | StructType | MapType | NumericType
        if isinstance(resolved, StructType):
            base = resolved
        elif isinstance(resolved, ArrayType):
            base = resolved
        elif isinstance(resolved, NumericType):
            base = resolved
        elif isinstance(resolved, MapType):
            base = resolved
        else:
            base = DataType(args[0].lower())
        if traits:
            for trait in traits:
                if trait not in self.environment.data_types:
                    raise ParseError(
                        f"Invalid trait (type) {trait} for {base}, line {meta.line}."
                    )
                matched = self.environment.data_types[trait]
                if not is_compatible_datatype(matched.type, base):
                    raise ParseError(
                        f"Invalid trait (type) {trait} for {base}, line {meta.line}. Trait expects type {matched.type}, has {base}"
                    )
            return TraitDataType(type=base, traits=traits)

        return base

    def array_comparison(self, args) -> ComparisonOperator:
        return ComparisonOperator([x.value.lower() for x in args])

    def COMPARISON_OPERATOR(self, args) -> ComparisonOperator:
        return ComparisonOperator(args.strip())

    def LOGICAL_OPERATOR(self, args) -> BooleanOperator:
        return BooleanOperator(args.lower())

    def concept_assignment(self, args):
        return args

    @v_args(meta=True)
    def column_assignment(self, meta: Meta, args):
        modifiers = []
        if len(args) == 2:
            alias = args[0]
            concept_list = args[1]
        else:
            alias = args[0][-1]
            concept_list = args[0]
        # recursively collect modifiers
        if len(concept_list) > 1:
            modifiers += concept_list[:-1]
        concept = concept_list[-1]
        resolved = self.environment.concepts.__getitem__(  # type: ignore
            key=concept, line_no=meta.line, file=self.token_address
        )
        return ColumnAssignment(
            alias=alias, modifiers=modifiers, concept=resolved.reference
        )

    def _TERMINATOR(self, args):
        return None

    def _static_functions(self, args):
        return args[0]

    def MODIFIER(self, args) -> Modifier:
        return Modifier(args.value)

    def SHORTHAND_MODIFIER(self, args) -> Modifier:
        return Modifier(args.value)

    def PURPOSE(self, args) -> Purpose:
        return Purpose(args.value)

    def AUTO(self, args) -> Purpose:
        return Purpose.AUTO

    def CONST(self, args) -> Purpose:
        return Purpose.CONSTANT

    def CONSTANT(self, args) -> Purpose:
        return Purpose.CONSTANT

    def PROPERTY(self, args):
        return Purpose.PROPERTY

    def HASH_TYPE(self, args):
        return args.value

    @v_args(meta=True)
    def prop_ident(self, meta: Meta, args) -> Tuple[List[Concept], str]:
        return [self.environment.concepts[grain] for grain in args[:-1]], args[-1]

    @v_args(meta=True)
    def concept_property_declaration(self, meta: Meta, args) -> Concept:
        unique = False
        if not args[0] == Purpose.PROPERTY:
            unique = True
            args = args[1:]
        metadata = Metadata()
        modifiers = []
        for arg in args:
            if isinstance(arg, Metadata):
                metadata = arg
            if isinstance(arg, Modifier):
                modifiers.append(arg)

        declaration = args[1]
        if isinstance(declaration, (tuple)):
            parents, name = declaration
            if "." in name:
                namespace, name = name.split(".", 1)
            else:
                namespace = self.environment.namespace or DEFAULT_NAMESPACE
        else:
            if "." not in declaration:
                raise ParseError(
                    f"Property declaration {args[1]} must be fully qualified with a parent key"
                )
            grain, name = declaration.rsplit(".", 1)
            parent = self.environment.concepts[grain]
            parents = [parent]
            namespace = parent.namespace
        concept = Concept(
            name=name,
            datatype=args[2],
            purpose=Purpose.PROPERTY if not unique else Purpose.UNIQUE_PROPERTY,
            metadata=metadata,
            grain=Grain(components={x.address for x in parents}),
            namespace=namespace,
            keys=set([x.address for x in parents]),
            modifiers=modifiers,
        )

        self.environment.add_concept(concept, meta)
        return concept

    @v_args(meta=True)
    def concept_declaration(self, meta: Meta, args) -> ConceptDeclarationStatement:
        metadata = Metadata()
        modifiers = []
        for arg in args:
            if isinstance(arg, Metadata):
                metadata = arg
            if isinstance(arg, Modifier):
                modifiers.append(arg)
        name = args[1]
        _, namespace, name, _ = parse_concept_reference(name, self.environment)
        concept = Concept(
            name=name,
            datatype=args[2],
            purpose=args[0],
            metadata=metadata,
            namespace=namespace,
            modifiers=modifiers,
            derivation=Derivation.ROOT,
            granularity=Granularity.MULTI_ROW,
        )
        if concept.metadata:
            concept.metadata.line_number = meta.line
        self.environment.add_concept(concept, meta=meta)
        return ConceptDeclarationStatement(concept=concept)

    @v_args(meta=True)
    def concept_derivation(self, meta: Meta, args) -> ConceptDerivationStatement:

        if len(args) > 3:
            metadata = args[3]
        else:
            metadata = None
        purpose = args[0]
        raw_name = args[1]
        # abc.def.property pattern
        if isinstance(raw_name, str):
            lookup, namespace, name, parent_concept = parse_concept_reference(
                raw_name, self.environment, purpose
            )
        # <abc.def,zef.gf>.property pattern
        else:
            keys, name = raw_name
            keys = [x.address for x in keys]
            namespaces = set([x.rsplit(".", 1)[0] for x in keys])
            if not len(namespaces) == 1:
                namespace = self.environment.namespace or DEFAULT_NAMESPACE
            else:
                namespace = namespaces.pop()
        source_value = args[2]
        # we need to strip off every parenthetical to see what is being assigned.
        while isinstance(source_value, Parenthetical):
            source_value = source_value.content

        if isinstance(
            source_value,
            (
                FilterItem,
                WindowItem,
                AggregateWrapper,
                Function,
                FunctionCallWrapper,
                Comparison,
            ),
        ):
            concept = arbitrary_to_concept(
                source_value,
                name=name,
                namespace=namespace,
                environment=self.environment,
                metadata=metadata,
            )

            # let constant purposes exist to support round-tripping
            # as a build concept may end up with a constant based on constant inlining happening recursively
            if purpose == Purpose.KEY and concept.purpose != Purpose.KEY:
                concept.purpose = Purpose.KEY
            elif (
                purpose
                and purpose != Purpose.AUTO
                and concept.purpose != purpose
                and purpose != Purpose.CONSTANT
            ):
                raise SyntaxError(
                    f'Concept {name} purpose {concept.purpose} does not match declared purpose {purpose}. Suggest defaulting to "auto"'
                )

            if concept.metadata:
                concept.metadata.line_number = meta.line
            self.environment.add_concept(concept, meta=meta)
            return ConceptDerivationStatement(concept=concept)

        elif isinstance(source_value, CONSTANT_TYPES):
            concept = constant_to_concept(
                source_value,
                name=name,
                namespace=namespace,
                metadata=metadata,
            )
            if concept.metadata:
                concept.metadata.line_number = meta.line
            self.environment.add_concept(concept, meta=meta)
            return ConceptDerivationStatement(concept=concept)

        raise SyntaxError(
            f"Received invalid type {type(args[2])} {args[2]} as input to concept derivation: `{self.text_lookup[self.token_address][meta.start_pos:meta.end_pos]}`"
        )

    @v_args(meta=True)
    def rowset_derivation_statement(
        self, meta: Meta, args
    ) -> RowsetDerivationStatement:
        name = args[0]
        select: SelectStatement | MultiSelectStatement = args[1]
        output = RowsetDerivationStatement(
            name=name,
            select=select,
            namespace=self.environment.namespace or DEFAULT_NAMESPACE,
        )

        for new_concept in rowset_to_concepts(output, self.environment):
            if new_concept.metadata:
                new_concept.metadata.line_number = meta.line
            self.environment.add_concept(new_concept, force=True)

        self.environment.add_rowset(
            output.name, output.select.as_lineage(self.environment)
        )
        return output

    @v_args(meta=True)
    def constant_derivation(self, meta: Meta, args) -> Concept:

        if len(args) > 3:
            metadata = args[3]
        else:
            metadata = None
        name = args[1]
        constant: Union[str, float, int, bool, MapWrapper, ListWrapper] = args[2]
        lookup, namespace, name, parent = parse_concept_reference(
            name, self.environment
        )
        concept = Concept(
            name=name,
            datatype=arg_to_datatype(constant),
            purpose=Purpose.CONSTANT,
            metadata=metadata,
            lineage=Function(
                operator=FunctionType.CONSTANT,
                output_datatype=arg_to_datatype(constant),
                output_purpose=Purpose.CONSTANT,
                arguments=[constant],
            ),
            grain=Grain(components=set()),
            namespace=namespace,
        )
        if concept.metadata:
            concept.metadata.line_number = meta.line
        self.environment.add_concept(concept, meta)
        return concept

    @v_args(meta=True)
    def concept(self, meta: Meta, args) -> ConceptDeclarationStatement:
        if isinstance(args[0], Concept):
            concept: Concept = args[0]
        else:
            concept = args[0].concept
        if concept.metadata:
            concept.metadata.line_number = meta.line
        return ConceptDeclarationStatement(concept=concept)

    def column_assignment_list(self, args):
        return args

    def column_list(self, args) -> List:
        return args

    def grain_clause(self, args) -> Grain:
        return Grain(
            components=set([self.environment.concepts[a].address for a in args[0]])
        )

    @v_args(meta=True)
    def aggregate_by(self, meta: Meta, args):
        base = args[0]
        b_concept = base.value.split(" ")[-1]
        args = [self.environment.concepts[a] for a in [b_concept] + args[1:]]
        return self.function_factory.create_function(args, FunctionType.GROUP, meta)

    def whole_grain_clause(self, args) -> WholeGrainWrapper:
        return WholeGrainWrapper(where=args[0])

    def MULTILINE_STRING(self, args) -> str:
        return args[3:-3]

    def raw_column_assignment(self, args):
        return RawColumnExpr(text=args[1])

    @v_args(meta=True)
    def datasource(self, meta: Meta, args):
        name = args[0]
        columns: List[ColumnAssignment] = args[1]
        grain: Optional[Grain] = None
        address: Optional[Address] = None
        where: Optional[WhereClause] = None
        non_partial_for: Optional[WhereClause] = None
        for val in args[1:]:
            if isinstance(val, Address):
                address = val
            elif isinstance(val, Grain):
                grain = val
            elif isinstance(val, WholeGrainWrapper):
                non_partial_for = val.where
            elif isinstance(val, Query):
                address = Address(location=f"({val.text})", is_query=True)
            elif isinstance(val, WhereClause):
                where = val
        if not address:
            raise ValueError(
                "Malformed datasource, missing address or query declaration"
            )
        datasource = Datasource(
            name=name,
            columns=columns,
            # grain will be set by default from args
            # TODO: move to factory
            grain=grain,  # type: ignore
            address=address,
            namespace=self.environment.namespace,
            where=where,
            non_partial_for=non_partial_for,
        )
        if datasource.where:
            for x in datasource.where.concept_arguments:
                if x.address not in datasource.output_concepts:
                    raise ValueError(
                        f"Datasource {name} where condition depends on concept {x.address} that does not exist on the datasource, line {meta.line}."
                    )
        if self.parse_pass == ParsePass.VALIDATION:
            self.environment.add_datasource(datasource, meta=meta)
            # if we have any foreign keys on the datasource, we can
            # at this point optimize them to properties if they do not have other usage.
            for column in columns:
                # skip partial for now
                if not grain:
                    continue
                if column.concept.address in grain.components:
                    continue
                target_c = self.environment.concepts[column.concept.address]
                if target_c.purpose != Purpose.KEY:
                    continue

                key_inputs = grain.components
                eligible = True
                for key in key_inputs:
                    # never overwrite a key with a dependency on a property
                    # for example - binding a datasource with a grain of <x>.fun should
                    # never override the grain of x to <fun>
                    if column.concept.address in (
                        self.environment.concepts[key].keys or set()
                    ):
                        eligible = False
                if not eligible:
                    continue
                keys = [self.environment.concepts[grain] for grain in key_inputs]
                # target_c.purpose = Purpose.PROPERTY
                target_c.keys = set([x.address for x in keys])
                # target_c.grain = Grain(components={x.address for x in keys})

        return datasource

    @v_args(meta=True)
    def comment(self, meta: Meta, args):
        assert len(args) == 1
        return Comment(text=args[0].value)

    def PARSE_COMMENT(self, args):
        return Comment(text=args.value.rstrip())

    @v_args(meta=True)
    def select_transform(self, meta: Meta, args) -> ConceptTransform:
        output: str = args[1]
        transformation = unwrap_transformation(args[0], self.environment)
        lookup, namespace, output, parent = parse_concept_reference(
            output, self.environment
        )

        metadata = Metadata(line_number=meta.line, concept_source=ConceptSource.SELECT)
        concept = arbitrary_to_concept(
            transformation,
            environment=self.environment,
            namespace=namespace,
            name=output,
            metadata=metadata,
        )
        return ConceptTransform(function=transformation, output=concept)

    @v_args(meta=True)
    def concept_nullable_modifier(self, meta: Meta, args) -> Modifier:
        return Modifier.NULLABLE

    @v_args(meta=True)
    def select_hide_modifier(self, meta: Meta, args) -> Modifier:
        return Modifier.HIDDEN

    @v_args(meta=True)
    def select_partial_modifier(self, meta: Meta, args) -> Modifier:
        return Modifier.PARTIAL

    @v_args(meta=True)
    def select_item(self, meta: Meta, args) -> Optional[SelectItem]:
        modifiers = [arg for arg in args if isinstance(arg, Modifier)]
        args = [arg for arg in args if not isinstance(arg, (Modifier, Comment))]

        if not args:
            return None
        if len(args) != 1:
            raise ParseError(
                "Malformed select statement"
                f" {args} {self.text_lookup[self.parse_address][meta.start_pos:meta.end_pos]}"
            )
        content = args[0]
        if isinstance(content, ConceptTransform):
            return SelectItem(content=content, modifiers=modifiers)
        return SelectItem(
            content=content,
            modifiers=modifiers,
        )

    def select_list(self, args):
        return [arg for arg in args if arg]

    def limit(self, args):
        return Limit(count=int(args[0].value))

    def ordering(self, args: list[str]):
        base = args[0].lower()
        if len(args) > 1:
            null_sort = args[-1]
            return Ordering(" ".join([base, "nulls", null_sort.lower()]))
        return Ordering(base)

    def order_list(self, args) -> List[OrderItem]:
        return [
            OrderItem(
                expr=x,
                order=y,
            )
            for x, y in zip(args[::2], args[1::2])
        ]

    def order_by(self, args):
        return OrderBy(items=args[0])

    def over_component(self, args):
        return ConceptRef(address=args[0].value.lstrip(",").strip())

    def over_list(self, args):
        return [x for x in args]

    def VALIDATION_SCOPE(self, args) -> ValidationScope:
        return ValidationScope(args.lower())

    @v_args(meta=True)
    def validate_statement(self, meta: Meta, args) -> ValidateStatement:
        if len(args) == 2:
            scope = args[0]
            targets = args[1].split(",")
        elif len(args) == 0:
            scope = ValidationScope.ALL
            targets = None
        else:
            scope = args[0]
            targets = None
        return ValidateStatement(
            scope=scope,
            targets=targets,
        )

    @v_args(meta=True)
    def merge_statement(self, meta: Meta, args) -> MergeStatementV2 | None:
        modifiers = []
        cargs: list[str] = []
        source_wildcard = None
        target_wildcard = None
        for arg in args:
            if isinstance(arg, Modifier):
                modifiers.append(arg)
            else:
                cargs.append(arg)
        source, target = cargs
        if source.endswith(".*"):
            if not target.endswith(".*"):
                raise ValueError("Invalid merge, source is wildcard, target is not")
            source_wildcard = source[:-2]
            target_wildcard = target[:-2]
            sources: list[Concept] = [
                v
                for k, v in self.environment.concepts.items()
                if v.namespace == source_wildcard
            ]
            targets: dict[str, Concept] = {}
            for x in sources:
                target = target_wildcard + "." + x.name
                if target in self.environment.concepts:
                    targets[x.address] = self.environment.concepts[target]
            sources = [x for x in sources if x.address in targets]
        else:
            sources = [self.environment.concepts[source]]
            targets = {sources[0].address: self.environment.concepts[target]}

        if self.parse_pass == ParsePass.VALIDATION:
            for source_c in sources:
                if isinstance(source_c, UndefinedConceptFull):
                    raise SyntaxError(
                        f"Cannot merge non-existent source concept {source_c.address} on line: {meta.line}"
                    )
            new = MergeStatementV2(
                sources=sources,
                targets=targets,
                modifiers=modifiers,
                source_wildcard=source_wildcard,
                target_wildcard=target_wildcard,
            )
            for source_c in new.sources:
                self.environment.merge_concept(
                    source_c, targets[source_c.address], modifiers
                )

            return new
        return None

    @v_args(meta=True)
    def rawsql_statement(self, meta: Meta, args) -> RawSQLStatement:
        statement = RawSQLStatement(meta=Metadata(line_number=meta.line), text=args[0])
        return statement

    def COPY_TYPE(self, args) -> IOType:
        return IOType(args.value)

    @v_args(meta=True)
    def copy_statement(self, meta: Meta, args) -> CopyStatement:
        return CopyStatement(
            target=args[1],
            target_type=args[0],
            meta=Metadata(line_number=meta.line),
            select=args[-1],
        )

    def resolve_import_address(self, address, is_stdlib: bool = False) -> str:
        if (
            isinstance(
                self.environment.config.import_resolver, FileSystemImportResolver
            )
            or is_stdlib
        ):
            with open(address, "r", encoding="utf-8") as f:
                text = f.read()
        elif isinstance(self.environment.config.import_resolver, DictImportResolver):
            lookup = address
            if lookup not in self.environment.config.import_resolver.content:
                raise ImportError(
                    f"Unable to import file {lookup}, not resolvable from provided source files."
                )
            text = self.environment.config.import_resolver.content[lookup]
        else:
            raise ImportError(
                f"Unable to import file {address}, resolver type {type(self.environment.config.import_resolver)} not supported"
            )
        return text

    def import_statement(self, args: list[str]) -> ImportStatement:
        start = datetime.now()
        is_file_resolver = isinstance(
            self.environment.config.import_resolver, FileSystemImportResolver
        )
        if len(args) == 2:
            alias = args[-1]
            cache_key = args[-1]
        else:
            alias = self.environment.namespace
            cache_key = args[0]
        input_path = args[0]
        path = input_path.split(".")
        is_stdlib = False
        if path[0] == "std":
            is_stdlib = True
            target = join(STDLIB_ROOT, *path) + ".preql"
            token_lookup: Path | str = Path(target)
        elif is_file_resolver:
            target = join(self.environment.working_path, *path) + ".preql"
            # tokens + text are cached by path
            token_lookup = Path(target)
        elif isinstance(self.environment.config.import_resolver, DictImportResolver):
            target = ".".join(path)
            token_lookup = target
        else:
            raise NotImplementedError

        # parser + env has to be cached by prior import path + current key
        key_path = self.import_keys + [cache_key]
        cache_lookup = "-".join(key_path)

        # we don't iterate past the max parse depth
        if len(key_path) > MAX_PARSE_DEPTH:
            return ImportStatement(
                alias=alias, input_path=input_path, path=Path(target)
            )

        if token_lookup in self.tokens:
            perf_logger.debug(f"\tTokens cached for {token_lookup}")
            raw_tokens = self.tokens[token_lookup]
            text = self.text_lookup[token_lookup]
        else:
            perf_logger.debug(f"\tTokens not cached for {token_lookup}, resolving")
            text = self.resolve_import_address(target, is_stdlib)
            self.text_lookup[token_lookup] = text

            try:
                raw_tokens = PARSER.parse(text)
            except Exception as e:
                raise ImportError(
                    f"Unable to import '{target}', parsing error: {e}"
                ) from e
            self.tokens[token_lookup] = raw_tokens

        if cache_lookup in self.parsed:
            perf_logger.debug(f"\tEnvironment cached for {token_lookup}")
            nparser = self.parsed[cache_lookup]
            new_env = nparser.environment
            if nparser.parse_pass != ParsePass.VALIDATION:
                # nparser.transform(raw_tokens)
                second_pass_start = datetime.now()
                nparser.run_second_parse_pass()
                second_pass_end = datetime.now()
                perf_logger.debug(
                    f"{second_pass_end - second_pass_start} seconds | Import {alias} key ({cache_key}) second pass took {second_pass_end - second_pass_start} to parse, {len(new_env.concepts)} concepts"
                )
        else:
            perf_logger.debug(f"\tParsing new for {token_lookup}")
            try:
                new_env = Environment(
                    working_path=dirname(target),
                    env_file_path=token_lookup,
                    config=self.environment.config,
                )
                new_env.concepts.fail_on_missing = False
                self.parsed[self.parse_address] = self
                nparser = ParseToObjects(
                    environment=new_env,
                    parse_address=cache_lookup,
                    token_address=token_lookup,
                    parsed=self.parsed,
                    tokens=self.tokens,
                    text_lookup=self.text_lookup,
                    import_keys=self.import_keys + [cache_key],
                    parse_config=self.parse_config,
                )
                nparser.transform(raw_tokens)
                self.parsed[cache_lookup] = nparser
            except Exception as e:
                raise ImportError(
                    f"Unable to import file {target}, parsing error: {e}"
                ) from e

        parsed_path = Path(args[0])
        imps = ImportStatement(alias=alias, input_path=input_path, path=parsed_path)

        self.environment.add_import(
            alias,
            new_env,
            Import(
                alias=alias,
                path=parsed_path,
                input_path=Path(target) if is_file_resolver else None,
            ),
        )
        end = datetime.now()
        perf_logger.debug(
            f"{end - start} seconds | Import {alias} key ({cache_key}) took  to parse, {len(new_env.concepts)} concepts"
        )
        return imps

    @v_args(meta=True)
    def show_category(self, meta: Meta, args) -> ShowCategory:
        return ShowCategory(args[0])

    @v_args(meta=True)
    def show_statement(self, meta: Meta, args) -> ShowStatement:
        return ShowStatement(content=args[0])

    @v_args(meta=True)
    def persist_statement(self, meta: Meta, args) -> PersistStatement | None:
        identifier: str = args[0]
        address: str = args[1]
        select: SelectStatement = args[2]
        if len(args) > 3:
            grain: Grain | None = args[3]
        else:
            grain = None
        if self.parse_pass == ParsePass.VALIDATION:
            new_datasource = select.to_datasource(
                namespace=(
                    self.environment.namespace
                    if self.environment.namespace
                    else DEFAULT_NAMESPACE
                ),
                name=identifier,
                address=Address(location=address),
                grain=grain,
                environment=self.environment,
            )
            return PersistStatement(
                select=select,
                datasource=new_datasource,
                meta=Metadata(line_number=meta.line),
            )
        return None

    @v_args(meta=True)
    def align_item(self, meta: Meta, args) -> AlignItem:
        return AlignItem(
            alias=args[0],
            namespace=self.environment.namespace,
            concepts=[self.environment.concepts[arg].reference for arg in args[1:]],
        )

    @v_args(meta=True)
    def align_clause(self, meta: Meta, args) -> AlignClause:
        return AlignClause(items=args)

    @v_args(meta=True)
    def derive_item(self, meta: Meta, args) -> DeriveItem:
        return DeriveItem(
            expr=args[0], name=args[1], namespace=self.environment.namespace
        )

    @v_args(meta=True)
    def derive_clause(self, meta: Meta, args) -> DeriveClause:

        return DeriveClause(items=args)

    @v_args(meta=True)
    def multi_select_statement(self, meta: Meta, args) -> MultiSelectStatement:

        selects: list[SelectStatement] = []
        align: AlignClause | None = None
        limit: int | None = None
        order_by: OrderBy | None = None
        where: WhereClause | None = None
        having: HavingClause | None = None
        derive: DeriveClause | None = None
        for arg in args:
            if isinstance(arg, SelectStatement):
                selects.append(arg)
            elif isinstance(arg, Limit):
                limit = arg.count
            elif isinstance(arg, OrderBy):
                order_by = arg
            elif isinstance(arg, WhereClause):
                where = arg
            elif isinstance(arg, HavingClause):
                having = arg
            elif isinstance(arg, AlignClause):
                align = arg
            elif isinstance(arg, DeriveClause):
                derive = arg

        assert align
        assert align is not None

        derived_concepts = []
        new_selects = [x.as_lineage(self.environment) for x in selects]
        lineage = MultiSelectLineage(
            selects=new_selects,
            align=align,
            derive=derive,
            namespace=self.environment.namespace,
            where_clause=where,
            having_clause=having,
            limit=limit,
            hidden_components=set(y for x in new_selects for y in x.hidden_components),
        )
        for x in align.items:
            concept = align_item_to_concept(
                x,
                align,
                selects,
                where=where,
                having=having,
                limit=limit,
                environment=self.environment,
            )
            derived_concepts.append(concept)
            self.environment.add_concept(concept, meta=meta)
        if derive:
            for derived in derive.items:
                derivation = derived.expr
                name = derived.name
                if not isinstance(derivation, (Function, Comparison, WindowItem)):
                    raise SyntaxError(
                        f"Invalid derive expression {derivation} in {meta.line}, must be a function or conditional"
                    )
                concept = derive_item_to_concept(
                    derivation, name, lineage, self.environment.namespace
                )
                derived_concepts.append(concept)
                self.environment.add_concept(concept, meta=meta)
        multi = MultiSelectStatement(
            selects=selects,
            align=align,
            namespace=self.environment.namespace,
            where_clause=where,
            order_by=order_by,
            limit=limit,
            meta=Metadata(line_number=meta.line),
            derived_concepts=derived_concepts,
            derive=derive,
        )
        return multi

    @v_args(meta=True)
    def select_statement(self, meta: Meta, args) -> SelectStatement:
        select_items: List[SelectItem] | None = None
        limit = None
        order_by = None
        where = None
        having = None
        for arg in args:
            if isinstance(arg, List):
                select_items = arg
            elif isinstance(arg, Limit):
                limit = arg.count
            elif isinstance(arg, OrderBy):
                order_by = arg
            elif isinstance(arg, WhereClause) and not isinstance(arg, HavingClause):
                where = arg
            elif isinstance(arg, HavingClause):
                having = arg
        if not select_items:
            raise ParseError("Malformed select, missing select items")
        pre_keys = set(self.environment.concepts.keys())
        base = SelectStatement.from_inputs(
            environment=self.environment,
            selection=select_items,
            order_by=order_by,
            where_clause=where,
            having_clause=having,
            limit=limit,
            meta=Metadata(line_number=meta.line),
        )
        if (
            self.parse_pass == ParsePass.INITIAL
            and self.parse_config.strict_name_shadow_enforcement
        ):
            intersection = base.locally_derived.intersection(pre_keys)
            if intersection:
                for x in intersection:
                    if str(base.local_concepts[x].lineage) == str(
                        self.environment.concepts[x].lineage
                    ):
                        local = base.local_concepts[x]
                        friendly_name = (
                            local.name
                            if local.namespace == DEFAULT_NAMESPACE
                            else local.namespace
                        )
                        raise NameShadowError(
                            f"Select statement {base} creates a new concept '{friendly_name}' with identical definition as the existing concept '{friendly_name}'. Replace {base.local_concepts[x].lineage} with a direct reference to {friendly_name}."
                        )
                else:
                    raise NameShadowError(
                        f"Select statement {base} creates new named concepts from calculations {list(intersection)} with identical name(s) to existing concept(s). Use new unique names for these."
                    )
        return base

    @v_args(meta=True)
    def address(self, meta: Meta, args):
        return args[0]

    @v_args(meta=True)
    def query(self, meta: Meta, args):
        return Query(text=args[0])

    def where(self, args):
        root = args[0]
        root = expr_to_boolean(root, self.function_factory)
        return WhereClause(conditional=root)

    def having(self, args):
        root = args[0]
        if not isinstance(root, (Comparison, Conditional, Parenthetical)):
            if arg_to_datatype(root) == DataType.BOOL:
                root = Comparison(left=root, right=True, operator=ComparisonOperator.EQ)
            else:
                root = Comparison(
                    left=root,
                    right=MagicConstants.NULL,
                    operator=ComparisonOperator.IS_NOT,
                )
        return HavingClause(conditional=root)

    @v_args(meta=True)
    def function_binding_list(self, meta: Meta, args) -> list[ArgBinding]:
        return args

    @v_args(meta=True)
    def function_binding_type(self, meta: Meta, args) -> FunctionBindingType:
        return FunctionBindingType(type=args[0])

    @v_args(meta=True)
    def function_binding_default(self, meta: Meta, args):
        return args[1]

    @v_args(meta=True)
    def function_binding_item(self, meta: Meta, args) -> ArgBinding:
        default = None
        type = None
        for arg in args[1:]:
            if isinstance(arg, FunctionBindingType):
                type = arg.type
            else:
                default = arg
        return ArgBinding.model_construct(name=args[0], datatype=type, default=default)

    @v_args(meta=True)
    def raw_function(self, meta: Meta, args) -> FunctionDeclaration:
        identity = args[0]
        function_arguments: list[ArgBinding] = args[1]
        output = args[2]

        self.environment.functions[identity] = CustomFunctionFactory(
            function=output,
            namespace=self.environment.namespace,
            function_arguments=function_arguments,
            name=identity,
        )
        return FunctionDeclaration(name=identity, args=function_arguments, expr=output)

    def custom_function(self, args) -> FunctionCallWrapper:
        name = args[0]
        args = args[1:]
        remapped = FunctionCallWrapper(
            content=self.environment.functions[name](*args), name=name, args=args
        )

        return remapped

    @v_args(meta=True)
    def function(self, meta: Meta, args) -> Function:
        return args[0]

    @v_args(meta=True)
    def type_drop_clause(self, meta: Meta, args) -> DropOn:
        return DropOn([FunctionType(x) for x in args])

    @v_args(meta=True)
    def type_add_clause(self, meta: Meta, args) -> AddOn:
        return AddOn([FunctionType(x) for x in args])

    @v_args(meta=True)
    def type_declaration(self, meta: Meta, args) -> TypeDeclaration:
        key = args[0]
        datatype: list[DataType] = [x for x in args[1:] if isinstance(x, DataType)]
        if len(datatype) == 1:
            final_datatype: list[DataType] | DataType = datatype[0]
        else:
            final_datatype = datatype
        add_on = None
        drop_on = None
        for x in args[1:]:
            if isinstance(x, AddOn):
                add_on = x
            elif isinstance(x, DropOn):
                drop_on = x
        new = CustomType(
            name=key,
            type=final_datatype,
            drop_on=drop_on.functions if drop_on else [],
            add_on=add_on.functions if add_on else [],
        )
        self.environment.data_types[key] = new
        return TypeDeclaration(type=new)

    def int_lit(self, args):
        return int("".join(args))

    def bool_lit(self, args):
        return args[0].capitalize() == "True"

    def null_lit(self, args):
        return NULL_VALUE

    def float_lit(self, args):
        return float(args[0])

    def array_lit(self, args):
        return list_to_wrapper(args)

    def tuple_lit(self, args):
        return tuple_to_wrapper(args)

    def string_lit(self, args) -> str:
        if not args:
            return ""

        return args[0]

    @v_args(meta=True)
    def struct_lit(self, meta, args):
        return self.function_factory.create_function(
            args, operator=FunctionType.STRUCT, meta=meta
        )

    def map_lit(self, args):
        parsed = dict(zip(args[::2], args[1::2]))
        wrapped = dict_to_map_wrapper(parsed)
        return wrapped

    def literal(self, args):
        return args[0]

    def product_operator(self, args) -> Function | Any:
        if len(args) == 1:
            return args[0]
        result = args[0]
        for i in range(1, len(args), 2):
            new_result = None
            op = args[i]
            right = args[i + 1]
            if op == "*":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.MULTIPLY
                )
            elif op == "/":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.DIVIDE
                )
            elif op == "%":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.MOD
                )
            else:
                raise ValueError(f"Unknown operator: {op}")
            result = new_result
        return new_result

    def PLUS_OR_MINUS(self, args) -> str:
        return args.value

    def MULTIPLY_DIVIDE_PERCENT(self, args) -> str:
        return args[0]

    @v_args(meta=True)
    def sum_operator(self, meta: Meta, args) -> Function | Any:
        if len(args) == 1:
            return args[0]
        result = args[0]
        for i in range(1, len(args), 2):
            new_result = None
            op = args[i]
            right = args[i + 1]
            if op == "+":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.ADD, meta=meta
                )
            elif op == "-":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.SUBTRACT, meta=meta
                )
            elif op == "||":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.CONCAT, meta=meta
                )
            elif op == "like":
                new_result = self.function_factory.create_function(
                    [result, right], operator=FunctionType.LIKE, meta=meta
                )
            else:
                raise ValueError(f"Unknown operator: {op}")
            result = new_result
        return result

    def comparison(self, args) -> Comparison:
        if len(args) == 1:
            return args[0]
        left = args[0]
        right = args[2]
        if args[1] in (ComparisonOperator.IN, ComparisonOperator.NOT_IN):
            return SubselectComparison(
                left=left,
                right=right,
                operator=args[1],
            )
        return Comparison(left=left, right=right, operator=args[1])

    def between_comparison(self, args) -> Conditional:
        left_bound = args[1]
        right_bound = args[2]
        return Conditional(
            left=Comparison(
                left=args[0], right=left_bound, operator=ComparisonOperator.GTE
            ),
            right=Comparison(
                left=args[0], right=right_bound, operator=ComparisonOperator.LTE
            ),
            operator=BooleanOperator.AND,
        )

    @v_args(meta=True)
    def subselect_comparison(self, meta: Meta, args) -> SubselectComparison:
        right = args[2]

        while isinstance(right, Parenthetical) and isinstance(
            right.content,
            (
                Concept,
                Function,
                FilterItem,
                WindowItem,
                AggregateWrapper,
                ListWrapper,
                TupleWrapper,
            ),
        ):
            right = right.content
        if isinstance(right, (Function, FilterItem, WindowItem, AggregateWrapper)):
            right_concept = arbitrary_to_concept(right, environment=self.environment)
            self.environment.add_concept(right_concept, meta=meta)
            right = right_concept.reference
        return SubselectComparison(
            left=args[0],
            right=right,
            operator=args[1],
        )

    def expr_tuple(self, args):
        datatypes = set([arg_to_datatype(x) for x in args])
        if len(datatypes) != 1:
            raise ParseError("Tuple must have same type for all elements")
        return TupleWrapper(val=tuple(args), type=datatypes.pop())

    def parenthetical(self, args):
        return Parenthetical(content=args[0])

    @v_args(meta=True)
    def condition_parenthetical(self, meta, args):
        if len(args) == 2:
            return Comparison(
                left=Parenthetical(content=args[1]),
                right=False,
                operator=ComparisonOperator.EQ,
            )
        return Parenthetical(content=args[0])

    def conditional(self, args):
        def munch_args(args):
            while args:
                if len(args) == 1:
                    return args[0]
                else:
                    return Conditional(
                        left=args[0], operator=args[1], right=munch_args(args[2:])
                    )

        return munch_args(args)

    def window_order(self, args):
        return WindowOrder(args[0])

    def window_order_by(self, args):
        # flatten tree
        return args[0]

    def window(self, args):

        return Window(count=args[1].value, window_order=args[0])

    def WINDOW_TYPE(self, args):
        return WindowType(args.strip())

    def window_item_over(self, args):

        return WindowItemOver(contents=args[0])

    def window_item_order(self, args):
        return WindowItemOrder(contents=args[0])

    def logical_operator(self, args):
        return BooleanOperator(args[0].value.lower())

    def DATE_PART(self, args):
        return DatePart(args.value)

    @v_args(meta=True)
    def window_item(self, meta: Meta, args) -> WindowItem:
        type: WindowType = args[0]
        order_by = []
        over = []
        index = None
        concept: Concept | None = None
        for item in args:
            if isinstance(item, int):
                index = item
            elif isinstance(item, WindowItemOrder):
                order_by = item.contents
            elif isinstance(item, WindowItemOver):
                over = item.contents
            elif isinstance(item, str):
                concept = self.environment.concepts[item]
            elif isinstance(item, ConceptRef):
                concept = self.environment.concepts[item.address]
            elif isinstance(item, WindowType):
                type = item
            else:
                concept = arbitrary_to_concept(item, environment=self.environment)
                self.environment.add_concept(concept, meta=meta)
        if not concept:
            raise ParseError(
                f"Window statements must be on fields, not constants - error in: `{self.text_lookup[self.parse_address][meta.start_pos:meta.end_pos]}`"
            )
        return WindowItem(
            type=type,
            content=concept.reference,
            over=over,
            order_by=order_by,
            index=index,
        )

    def filter_item(self, args) -> FilterItem:
        where: WhereClause
        expr, raw = args
        if isinstance(raw, WhereClause):
            where = raw
        else:
            where = WhereClause(conditional=expr_to_boolean(raw, self.function_factory))
        if isinstance(expr, str):
            expr = self.environment.concepts[expr].reference
        return FilterItem(content=expr, where=where)

    # BEGIN FUNCTIONS
    @v_args(meta=True)
    def expr_reference(self, meta, args) -> Concept:
        return self.environment.concepts.__getitem__(args[0], meta.line)

    def expr(self, args):
        if len(args) > 1:
            raise ParseError("Expression should have one child only.")
        return args[0]

    def aggregate_over(self, args):
        return args[0]

    def aggregate_all(self, args):
        return [
            ConceptRef(
                address=f"{INTERNAL_NAMESPACE}.{ALL_ROWS_CONCEPT}",
                datatype=DataType.INTEGER,
            )
        ]

    def aggregate_functions(self, args):
        if len(args) == 2:
            return AggregateWrapper(function=args[0], by=args[1])
        return AggregateWrapper(function=args[0])

    @v_args(meta=True)
    def index_access(self, meta, args):
        args = process_function_args(args, meta=meta, environment=self.environment)
        base = args[0]
        if base.datatype == DataType.MAP or isinstance(base.datatype, MapType):
            return self.function_factory.create_function(
                args, FunctionType.MAP_ACCESS, meta
            )
        return self.function_factory.create_function(
            args, FunctionType.INDEX_ACCESS, meta
        )

    @v_args(meta=True)
    def map_key_access(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.MAP_ACCESS, meta
        )

    @v_args(meta=True)
    def attr_access(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.ATTR_ACCESS, meta
        )

    @v_args(meta=True)
    def fcoalesce(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.COALESCE, meta)

    @v_args(meta=True)
    def fnullif(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.NULLIF, meta)

    @v_args(meta=True)
    def frecurse_edge(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.RECURSE_EDGE, meta
        )

    @v_args(meta=True)
    def unnest(self, meta, args):

        return self.function_factory.create_function(args, FunctionType.UNNEST, meta)

    @v_args(meta=True)
    def count(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.COUNT, meta)

    @v_args(meta=True)
    def fgroup(self, meta, args):
        if len(args) == 2:
            fargs = [args[0]] + list(args[1])
        else:
            fargs = [args[0]]
        return self.function_factory.create_function(fargs, FunctionType.GROUP, meta)

    @v_args(meta=True)
    def fabs(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.ABS, meta)

    @v_args(meta=True)
    def count_distinct(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.COUNT_DISTINCT, meta
        )

    @v_args(meta=True)
    def sum(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.SUM, meta)

    @v_args(meta=True)
    def array_agg(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.ARRAY_AGG, meta)

    @v_args(meta=True)
    def any(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.ANY, meta)

    @v_args(meta=True)
    def avg(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.AVG, meta)

    @v_args(meta=True)
    def max(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.MAX, meta)

    @v_args(meta=True)
    def min(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.MIN, meta)

    @v_args(meta=True)
    def len(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.LENGTH, meta)

    @v_args(meta=True)
    def fsplit(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.SPLIT, meta)

    @v_args(meta=True)
    def concat(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.CONCAT, meta)

    @v_args(meta=True)
    def union(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.UNION, meta)

    @v_args(meta=True)
    def like(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.LIKE, meta)

    @v_args(meta=True)
    def alt_like(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.LIKE, meta)

    @v_args(meta=True)
    def ilike(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.LIKE, meta)

    @v_args(meta=True)
    def upper(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.UPPER, meta)

    @v_args(meta=True)
    def fstrpos(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.STRPOS, meta)

    @v_args(meta=True)
    def freplace(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.REPLACE, meta)

    @v_args(meta=True)
    def fcontains(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.CONTAINS, meta)

    @v_args(meta=True)
    def ftrim(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.TRIM, meta)

    @v_args(meta=True)
    def fhash(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.HASH, meta)

    @v_args(meta=True)
    def fsubstring(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.SUBSTRING, meta)

    @v_args(meta=True)
    def flower(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.LOWER, meta)

    @v_args(meta=True)
    def fregexp_contains(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.REGEXP_CONTAINS, meta
        )

    @v_args(meta=True)
    def fregexp_extract(self, meta, args):
        if len(args) == 2:
            # this is a magic value to represent the default behavior
            args.append(-1)
        return self.function_factory.create_function(
            args, FunctionType.REGEXP_EXTRACT, meta
        )

    @v_args(meta=True)
    def fregexp_replace(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.REGEXP_REPLACE, meta
        )

    # date functions
    @v_args(meta=True)
    def fdate(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATE, meta)

    @v_args(meta=True)
    def fdate_trunc(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.DATE_TRUNCATE, meta
        )

    @v_args(meta=True)
    def fdate_part(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATE_PART, meta)

    @v_args(meta=True)
    def fdate_add(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATE_ADD, meta)

    @v_args(meta=True)
    def fdate_sub(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATE_SUB, meta)

    @v_args(meta=True)
    def fdate_diff(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATE_DIFF, meta)

    @v_args(meta=True)
    def fdatetime(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DATETIME, meta)

    @v_args(meta=True)
    def ftimestamp(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.TIMESTAMP, meta)

    @v_args(meta=True)
    def fsecond(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.SECOND, meta)

    @v_args(meta=True)
    def fminute(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.MINUTE, meta)

    @v_args(meta=True)
    def fhour(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.HOUR, meta)

    @v_args(meta=True)
    def fday(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DAY, meta)

    @v_args(meta=True)
    def fday_name(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.DAY_NAME, meta)

    @v_args(meta=True)
    def fday_of_week(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.DAY_OF_WEEK, meta
        )

    @v_args(meta=True)
    def fweek(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.WEEK, meta)

    @v_args(meta=True)
    def fmonth(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.MONTH, meta)

    @v_args(meta=True)
    def fmonth_name(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.MONTH_NAME, meta
        )

    @v_args(meta=True)
    def fquarter(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.QUARTER, meta)

    @v_args(meta=True)
    def fyear(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.YEAR, meta)

    def internal_fcast(self, meta, args) -> Function:
        args = process_function_args(args, meta=meta, environment=self.environment)

        if isinstance(args[0], str):

            processed: date | datetime | int | float | bool | str
            if args[1] == DataType.DATE:
                processed = date.fromisoformat(args[0])
            elif args[1] == DataType.DATETIME:
                processed = datetime.fromisoformat(args[0])
            elif args[1] == DataType.TIMESTAMP:
                processed = datetime.fromisoformat(args[0])
            elif args[1] == DataType.INTEGER:
                processed = int(args[0])
            elif args[1] == DataType.FLOAT:
                processed = float(args[0])
            elif args[1] == DataType.BOOL:
                processed = args[0].capitalize() == "True"
            elif args[1] == DataType.STRING:
                processed = args[0]
            else:
                raise SyntaxError(f"Invalid cast type {args[1]}")
            if isinstance(args[1], TraitDataType):
                return self.function_factory.create_function(
                    [processed, args[1]], FunctionType.TYPED_CONSTANT, meta
                )
            return self.function_factory.create_function(
                [processed], FunctionType.CONSTANT, meta
            )
        return self.function_factory.create_function(args, FunctionType.CAST, meta)

    @v_args(meta=True)
    def fdate_spine(self, meta, args) -> Function:
        return self.function_factory.create_function(
            args, FunctionType.DATE_SPINE, meta
        )

    # utility functions
    @v_args(meta=True)
    def fcast(self, meta, args) -> Function:
        return self.internal_fcast(meta, args)

    # math functions
    @v_args(meta=True)
    def fadd(self, meta, args) -> Function:

        return self.function_factory.create_function(args, FunctionType.ADD, meta)

    @v_args(meta=True)
    def fsub(self, meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.SUBTRACT, meta)

    @v_args(meta=True)
    def fmul(self, meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.MULTIPLY, meta)

    @v_args(meta=True)
    def fdiv(self, meta: Meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.DIVIDE, meta)

    @v_args(meta=True)
    def fmod(self, meta: Meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.MOD, meta)

    @v_args(meta=True)
    def fsqrt(self, meta: Meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.SQRT, meta)

    @v_args(meta=True)
    def frandom(self, meta: Meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.RANDOM, meta)

    @v_args(meta=True)
    def fround(self, meta, args) -> Function:
        if len(args) == 1:
            args.append(0)
        return self.function_factory.create_function(args, FunctionType.ROUND, meta)

    @v_args(meta=True)
    def flog(self, meta, args) -> Function:
        if len(args) == 1:
            args.append(10)
        return self.function_factory.create_function(args, FunctionType.LOG, meta)

    @v_args(meta=True)
    def ffloor(self, meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.FLOOR, meta)

    @v_args(meta=True)
    def fceil(self, meta, args) -> Function:
        return self.function_factory.create_function(args, FunctionType.CEIL, meta)

    @v_args(meta=True)
    def fcase(self, meta, args: List[Union[CaseWhen, CaseElse]]) -> Function:
        return self.function_factory.create_function(args, FunctionType.CASE, meta)

    @v_args(meta=True)
    def fcase_when(self, meta, args) -> CaseWhen:
        args = process_function_args(args, meta=meta, environment=self.environment)
        root = expr_to_boolean(args[0], self.function_factory)
        return CaseWhen(comparison=root, expr=args[1])

    @v_args(meta=True)
    def fcase_else(self, meta, args) -> CaseElse:
        args = process_function_args(args, meta=meta, environment=self.environment)
        return CaseElse(expr=args[0])

    @v_args(meta=True)
    def fcurrent_date(self, meta, args):
        return CurrentDate([])

    @v_args(meta=True)
    def fcurrent_datetime(self, meta, args):
        return self.function_factory.create_function(
            args=[], operator=FunctionType.CURRENT_DATETIME, meta=meta
        )

    @v_args(meta=True)
    def fcurrent_timestamp(self, meta, args):
        return self.function_factory.create_function(
            args=[], operator=FunctionType.CURRENT_TIMESTAMP, meta=meta
        )

    @v_args(meta=True)
    def fnot(self, meta, args):
        if arg_to_datatype(args[0]) == DataType.BOOL:
            return Comparison(
                left=self.function_factory.create_function(
                    [args[0], False], FunctionType.COALESCE, meta
                ),
                operator=ComparisonOperator.EQ,
                right=False,
                meta=meta,
            )
        return self.function_factory.create_function(args, FunctionType.IS_NULL, meta)

    @v_args(meta=True)
    def fbool(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.BOOL, meta)

    @v_args(meta=True)
    def fmap_keys(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.MAP_KEYS, meta)

    @v_args(meta=True)
    def fmap_values(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.MAP_VALUES, meta
        )

    @v_args(meta=True)
    def farray_sum(self, meta, args):
        return self.function_factory.create_function(args, FunctionType.ARRAY_SUM, meta)

    @v_args(meta=True)
    def farray_distinct(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.ARRAY_DISTINCT, meta
        )

    @v_args(meta=True)
    def farray_to_string(self, meta, args):
        return self.function_factory.create_function(
            args, FunctionType.ARRAY_TO_STRING, meta
        )

    @v_args(meta=True)
    def farray_sort(self, meta, args):
        if len(args) == 1:
            # this is a magic value to represent the default behavior
            args = [args[0], Ordering.ASCENDING]
        return self.function_factory.create_function(
            args, FunctionType.ARRAY_SORT, meta
        )

    @v_args(meta=True)
    def transform_lambda(self, meta, args):
        return self.environment.functions[args[0]]

    @v_args(meta=True)
    def farray_transform(self, meta, args) -> Function:
        factory: CustomFunctionFactory = args[1]
        if not len(factory.function_arguments) == 1:
            raise InvalidSyntaxException(
                "Array transform function must have exactly one argument;"
            )
        array_type = arg_to_datatype(args[0])
        if not isinstance(array_type, ArrayType):
            raise InvalidSyntaxException(
                f"Array transform function must be applied to an array, not {array_type}"
            )
        return self.function_factory.create_function(
            [
                args[0],
                factory.function_arguments[0],
                factory(
                    ArgBinding(
                        name=factory.function_arguments[0].name,
                        datatype=array_type.value_data_type,
                    )
                ),
            ],
            FunctionType.ARRAY_TRANSFORM,
            meta,
        )

    @v_args(meta=True)
    def farray_filter(self, meta, args) -> Function:
        factory: CustomFunctionFactory = args[1]
        if not len(factory.function_arguments) == 1:
            raise InvalidSyntaxException(
                "Array filter function must have exactly one argument;"
            )
        array_type = arg_to_datatype(args[0])
        if not isinstance(array_type, ArrayType):
            raise InvalidSyntaxException(
                f"Array filter function must be applied to an array, not {array_type}"
            )
        return self.function_factory.create_function(
            [
                args[0],
                factory.function_arguments[0],
                factory(
                    ArgBinding(
                        name=factory.function_arguments[0].name,
                        datatype=array_type.value_data_type,
                    )
                ),
            ],
            FunctionType.ARRAY_FILTER,
            meta,
        )


def unpack_visit_error(e: VisitError, text: str | None = None):
    """This is required to get exceptions from imports, which would
    raise nested VisitErrors"""
    if isinstance(e.orig_exc, VisitError):
        unpack_visit_error(e.orig_exc, text)
    elif isinstance(e.orig_exc, (UndefinedConceptException, ImportError)):
        raise e.orig_exc
    elif isinstance(e.orig_exc, (SyntaxError, TypeError)):
        if isinstance(e.obj, Tree):
            if text:
                extract = text[e.obj.meta.start_pos - 5 : e.obj.meta.end_pos + 5]
                raise InvalidSyntaxException(
                    str(e.orig_exc)
                    + " Raised when parsing rule: "
                    + str(e.rule)
                    + f' Line: {e.obj.meta.line} "...{extract}..."'
                )
            InvalidSyntaxException(
                str(e.orig_exc) + " in " + str(e.rule) + f" Line: {e.obj.meta.line}"
            )
        raise InvalidSyntaxException(str(e.orig_exc)).with_traceback(
            e.orig_exc.__traceback__
        )
    raise e.orig_exc


def parse_text_raw(text: str, environment: Optional[Environment] = None):
    PARSER.parse(text)


ERROR_CODES: dict[int, str] = {
    # 100 code are SQL compatability errors
    101: "Using FROM keyword? Trilogy does not have a FROM clause (Datasource resolution is automatic).",
    # 200 codes relate to required explicit syntax (we could loosen these?)
    201: 'Missing alias? Alias must be specified with "AS" - e.g. `SELECT x+1 AS y`',
    202: "Missing closing semicolon? Statements must be terminated with a semicolon `;`.",
    210: "Missing order direction? Order by must be explicit about direction - specify `asc` or `desc`.",
}

DEFAULT_ERROR_SPAN: int = 30


def inject_context_maker(pos: int, text: str, span: int = 40) -> str:
    """Returns a pretty string pinpointing the error in the text,
    with span amount of context characters around it.

    Note:
        The parser doesn't hold a copy of the text it has to parse,
        so you have to provide it again
    """

    start = max(pos - span, 0)
    end = pos + span
    if not isinstance(text, bytes):

        before = text[start:pos].rsplit("\n", 1)[-1]
        after = text[pos:end].split("\n", 1)[0]
        rcap = ""
        # if it goes beyond the end of text, no ...
        # if it terminates on a space, no need for ...
        if after and not after[-1].isspace() and not (end > len(text)):
            rcap = "..."
        lcap = ""
        if start > 0 and not before[0].isspace():
            lcap = "..."
        lpad = " "
        rpad = " "
        if before.endswith(" "):
            lpad = ""
        if after.startswith(" "):
            rpad = ""
        return f"{lcap}{before}{lpad}???{rpad}{after}{rcap}"
    else:
        before = text[start:pos].rsplit(b"\n", 1)[-1]
        after = text[pos:end].split(b"\n", 1)[0]
        return (before + b" ??? " + after).decode("ascii", "backslashreplace")


def parse_text(
    text: str,
    environment: Optional[Environment] = None,
    root: Path | None = None,
    parse_config: Parsing | None = None,
) -> Tuple[
    Environment,
    List[
        Datasource
        | ImportStatement
        | SelectStatement
        | PersistStatement
        | ShowStatement
        | RawSQLStatement
        | ValidateStatement
        | None
    ],
]:
    def _create_syntax_error(code: int, pos: int, text: str) -> InvalidSyntaxException:
        """Helper to create standardized syntax error with context."""
        return InvalidSyntaxException(
            f"Syntax [{code}]: "
            + ERROR_CODES[code]
            + "\nLocation:\n"
            + inject_context_maker(pos, text.replace("\n", " "), DEFAULT_ERROR_SPAN)
        )

    def _create_generic_syntax_error(
        message: str, pos: int, text: str
    ) -> InvalidSyntaxException:
        """Helper to create generic syntax error with context."""
        return InvalidSyntaxException(
            message
            + "\nLocation:\n"
            + inject_context_maker(pos, text.replace("\n", " "), DEFAULT_ERROR_SPAN)
        )

    def _handle_unexpected_token(e: UnexpectedToken, text: str) -> None:
        """Handle UnexpectedToken errors to make friendlier error messages."""
        # Handle ordering direction error
        pos = e.pos_in_stream or 0
        if e.interactive_parser.lexer_thread.state:
            last_token = e.interactive_parser.lexer_thread.state.last_token
        else:
            last_token = None
        if e.expected == {"ORDERING_DIRECTION"}:
            raise _create_syntax_error(210, pos, text)

        # Handle FROM token error
        parsed_tokens = (
            [x.value for x in e.token_history if x] if e.token_history else []
        )

        if parsed_tokens == ["FROM"]:
            raise _create_syntax_error(101, pos, text)
        # check if they are missing a semicolon
        if last_token and e.token.type == "$END":
            try:

                e.interactive_parser.feed_token(Token("_TERMINATOR", ";"))
                state = e.interactive_parser.lexer_thread.state
                if state and state.last_token:
                    new_pos = state.last_token.end_pos or pos
                else:
                    new_pos = pos
                raise _create_syntax_error(202, new_pos, text)
            except UnexpectedToken:
                pass
        # check if they forgot an as
        try:
            e.interactive_parser.feed_token(Token("AS", "AS"))
            state = e.interactive_parser.lexer_thread.state
            if state and state.last_token:
                new_pos = state.last_token.end_pos or pos
            else:
                new_pos = pos
            e.interactive_parser.feed_token(Token("IDENTIFIER", e.token.value))
            raise _create_syntax_error(201, new_pos, text)
        except UnexpectedToken:
            pass

        # Default UnexpectedToken handling
        raise _create_generic_syntax_error(str(e), pos, text)

    environment = environment or (
        Environment(working_path=root) if root else Environment()
    )
    parser = ParseToObjects(
        environment=environment, import_keys=["root"], parse_config=parse_config
    )
    start = datetime.now()

    try:
        parser.set_text(text)
        # disable fail on missing to allow for circular dependencies
        parser.prepare_parse()
        parser.transform(PARSER.parse(text))
        # this will reset fail on missing
        pass_two = parser.run_second_parse_pass()
        output = [v for v in pass_two if v]
        environment.concepts.fail_on_missing = True
        end = datetime.now()
        perf_logger.debug(
            f"Parse time: {end - start} for {len(text)} characters, {len(output)} objects"
        )
    except VisitError as e:
        unpack_visit_error(e, text)
        # this will never be reached
        raise e
    except UnexpectedToken as e:
        _handle_unexpected_token(e, text)
    except (UnexpectedCharacters, UnexpectedEOF, UnexpectedInput) as e:
        raise _create_generic_syntax_error(str(e), e.pos_in_stream or 0, text)
    except (ValidationError, TypeError) as e:
        raise InvalidSyntaxException(str(e))

    return environment, output



import sys
from collections import defaultdict
from datetime import date, datetime, timedelta
from typing import List, Tuple, TypeVar

from trilogy.core.enums import ComparisonOperator, FunctionType
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildFunction,
    BuildParenthetical,
)
from trilogy.core.models.core import DataType

# Define a generic type that ensures start and end are the same type
T = TypeVar("T", int, float, date, datetime)


def reduce_expression(
    var: BuildConcept, group_tuple: list[tuple[ComparisonOperator, T]]
) -> bool:
    # Track ranges
    lower_check: T
    upper_check: T

    # if var.datatype in (DataType.FLOAT,):
    #     lower_check = float("-inf")  # type: ignore
    #     upper_check = float("inf")  # type: ignore
    if var.datatype == DataType.INTEGER:
        lower_check = float("-inf")  # type: ignore
        upper_check = float("inf")  # type: ignore
    elif var.datatype == DataType.DATE:
        lower_check = date.min  # type: ignore
        upper_check = date.max  # type: ignore

    elif var.datatype == DataType.DATETIME:
        lower_check = datetime.min  # type: ignore
        upper_check = datetime.max  # type: ignore
    elif var.datatype == DataType.BOOL:
        lower_check = False  # type: ignore
        upper_check = True  # type: ignore
    elif var.datatype == DataType.FLOAT:
        lower_check = float("-inf")  # type: ignore
        upper_check = float("inf")  # type: ignore
    else:
        return False

    ranges: list[Tuple[T, T]] = []
    for op, value in group_tuple:
        increment: int | timedelta | float
        if isinstance(value, date):
            increment = timedelta(days=1)
        elif isinstance(value, datetime):
            increment = timedelta(seconds=1)
        elif isinstance(value, int):
            increment = 1
        elif isinstance(value, float):
            increment = sys.float_info.epsilon

        if op == ">":
            ranges.append(
                (
                    value + increment,
                    upper_check,
                )
            )
        elif op == ">=":
            ranges.append(
                (
                    value,
                    upper_check,
                )
            )
        elif op == "<":
            ranges.append(
                (
                    lower_check,
                    value - increment,
                )
            )
        elif op == "<=":
            ranges.append(
                (
                    lower_check,
                    value,
                )
            )
        elif op == "=":
            ranges.append(
                (
                    value,
                    value,
                )
            )
        elif op == ComparisonOperator.IS:
            ranges.append(
                (
                    value,
                    value,
                )
            )
        elif op == ComparisonOperator.NE:
            pass
        else:
            return False
    return is_fully_covered(lower_check, upper_check, ranges, increment)


TARGET_TYPES = (
    int,
    date,
    float,
    datetime,
    bool,
)
REDUCABLE_TYPES = (int, float, date, bool, datetime, BuildFunction)


def simplify_conditions(
    conditions: list[BuildComparison | BuildConditional | BuildParenthetical],
) -> bool:
    # Group conditions by variable
    grouped: dict[
        BuildConcept, list[tuple[ComparisonOperator, datetime | int | date | float]]
    ] = defaultdict(list)
    for condition in conditions:
        if not isinstance(condition, BuildComparison):
            return False
        left_is_concept = False
        left_is_reducable = False
        right_is_concept = False
        right_is_reducable = False
        if isinstance(condition.left, BuildConcept):
            left_is_concept = True
        elif isinstance(condition.left, REDUCABLE_TYPES):
            left_is_reducable = True

        if isinstance(condition.right, BuildConcept):
            right_is_concept = True
        elif isinstance(condition.right, REDUCABLE_TYPES):
            right_is_reducable = True

        if not (
            (left_is_concept and right_is_reducable)
            or (right_is_concept and left_is_reducable)
        ):
            return False
        if left_is_concept:
            concept = condition.left
            raw_comparison = condition.right
        else:
            concept = condition.right
            raw_comparison = condition.left

        if isinstance(raw_comparison, BuildFunction):
            if not raw_comparison.operator == FunctionType.CONSTANT:
                return False
            first_arg = raw_comparison.arguments[0]
            if not isinstance(first_arg, TARGET_TYPES):
                return False
            comparison = first_arg
        else:
            if not isinstance(raw_comparison, TARGET_TYPES):
                return False
            comparison = raw_comparison

        if not isinstance(comparison, REDUCABLE_TYPES):
            return False

        var: BuildConcept = concept  # type: ignore
        op = condition.operator
        grouped[var].append((op, comparison))

    simplified = []
    for var, group_tuple in grouped.items():
        simplified.append(reduce_expression(var, group_tuple))  # type: ignore

    # Final simplification
    return True if all(isinstance(s, bool) and s for s in simplified) else False


def boolean_fully_covered(
    start: bool,
    end: bool,
    ranges: List[Tuple[bool, bool]],
):
    all = []
    for r_start, r_end in ranges:
        if r_start is True and r_end is True:
            all.append(True)
        elif r_start is False and r_end is False:
            all.append(False)
    return set(all) == {False, True}


def is_fully_covered(
    start: T,
    end: T,
    ranges: List[Tuple[T, T]],
    increment: int | timedelta | float,
):
    """
    Check if the list of range pairs fully covers the set [start, end].

    Parameters:
    - start (int or float): The starting value of the set to cover.
    - end (int or float): The ending value of the set to cover.
    - ranges (list of tuples): List of range pairs [(start1, end1), (start2, end2), ...].

    Returns:
    - bool: True if the ranges fully cover [start, end], False otherwise.
    """
    if isinstance(start, bool) and isinstance(end, bool):
        # convert each element of each tuple to a boolean
        bool_ranges = [(bool(r_start), bool(r_end)) for r_start, r_end in ranges]

        return boolean_fully_covered(start, end, bool_ranges)
    # Sort ranges by their start values (and by end values for ties)
    ranges.sort()

    # Check for gaps
    current_end = start
    for r_start, r_end in ranges:
        # If there's a gap between the current range and the previous coverage
        if (r_start - current_end) > increment:  # type: ignore
            return False
        # Extend the current coverage
        current_end = max(current_end, r_end)

    # If the loop ends and we haven't reached the end, return False
    return current_end >= end


def get_union_sources(
    datasources: list[BuildDatasource], concepts: list[BuildConcept]
) -> List[list[BuildDatasource]]:
    candidates: list[BuildDatasource] = []

    for x in datasources:
        if any([c.address in x.output_concepts for c in concepts]):
            if (
                any([c.address in x.partial_concepts for c in concepts])
                and x.non_partial_for
            ):
                candidates.append(x)
    assocs: dict[str, list[BuildDatasource]] = defaultdict(list[BuildDatasource])
    for x in candidates:
        if not x.non_partial_for:
            continue
        if not len(x.non_partial_for.concept_arguments) == 1:
            continue
        merge_key = x.non_partial_for.concept_arguments[0]
        assocs[merge_key.address].append(x)
    final: list[list[BuildDatasource]] = []
    for _, dses in assocs.items():
        conditions = [c.non_partial_for.conditional for c in dses if c.non_partial_for]
        if simplify_conditions(conditions):
            final.append(dses)
    return final



from itertools import combinations
from typing import Callable, List, Optional

import networkx as nx
from networkx.algorithms import approximation as ax

from trilogy.constants import logger
from trilogy.core.enums import Derivation, FunctionType
from trilogy.core.exceptions import AmbiguousRelationshipResolutionException
from trilogy.core.graph_models import (
    ReferenceGraph,
    concept_to_node,
    prune_sources_for_conditions,
)
from trilogy.core.models.build import (
    BuildConcept,
    BuildConditional,
    BuildFunction,
    BuildGrain,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import History, MergeNode, StrategyNode
from trilogy.core.processing.utility import padding
from trilogy.utility import unique

LOGGER_PREFIX = "[GEN_MERGE_NODE]"
AMBIGUITY_CHECK_LIMIT = 20


def filter_pseudonyms_for_source(
    ds_graph: nx.DiGraph, node: str, pseudonyms: set[tuple[str, str]]
):
    to_remove = set()
    for edge in ds_graph.edges:
        if edge in pseudonyms:
            lengths = {}
            for n in edge:
                try:
                    lengths[n] = nx.shortest_path_length(ds_graph, node, n)
                except nx.NetworkXNoPath:
                    lengths[n] = 999
            to_remove.add(max(lengths, key=lambda x: lengths.get(x, 0)))
    for node in to_remove:
        ds_graph.remove_node(node)


def extract_address(node: str):
    return node.split("~")[1].split("@")[0]


def extract_concept(node: str, env: BuildEnvironment):
    if node in env.alias_origin_lookup:
        return env.alias_origin_lookup[node]
    return env.concepts[node]


def filter_unique_graphs(graphs: list[list[str]]) -> list[list[str]]:
    unique_graphs: list[set[str]] = []

    # sort graphs from largest to smallest
    graphs.sort(key=lambda x: len(x), reverse=True)
    for graph in graphs:
        if not any(set(graph).issubset(x) for x in unique_graphs):
            unique_graphs.append(set(graph))

    return [list(x) for x in unique_graphs]


def extract_ds_components(
    g: nx.DiGraph, nodelist: list[str], pseudonyms: set[tuple[str, str]]
) -> list[list[str]]:
    graphs = []
    for node in g.nodes:
        if node.startswith("ds~"):
            local = g.copy()
            filter_pseudonyms_for_source(local, node, pseudonyms)
            ds_graph: nx.DiGraph = nx.ego_graph(local, node, radius=10).copy()
            graphs.append(
                [
                    extract_address(x)
                    for x in ds_graph.nodes
                    if not str(x).startswith("ds~")
                ]
            )
    # if we had no ego graphs, return all concepts
    if not graphs:
        return [[extract_address(node) for node in nodelist]]
    graphs = filter_unique_graphs(graphs)
    for node in nodelist:
        parsed = extract_address(node)
        if not any(parsed in x for x in graphs):
            graphs.append([parsed])
    return graphs


def prune_and_merge(
    G: nx.DiGraph,
    keep_node_lambda: Callable[[str], bool],
) -> nx.DiGraph:
    """
    Prune nodes of one type and create direct connections between remaining nodes.

    Args:
        G: NetworkX graph
        keep_node_type: The node type to keep
        node_type_attr: Attribute name that stores node type

    Returns:
        New graph with only keep_node_type nodes and merged connections
    """
    # Get nodes to keep
    nodes_to_keep = [n for n in G.nodes if keep_node_lambda(n)]
    # Create new graph
    new_graph = G.subgraph(nodes_to_keep).copy()

    # Find paths between nodes to keep through removed nodes
    nodes_to_remove = [n for n in G.nodes() if n not in nodes_to_keep]

    for node_pair in combinations(nodes_to_keep, 2):
        n1, n2 = node_pair

        # Check if there's a path through removed nodes
        try:
            path = nx.shortest_path(G, n1, n2)
            # If path exists and goes through nodes we're removing
            if len(path) > 2 or any(node in nodes_to_remove for node in path[1:-1]):
                new_graph.add_edge(n1, n2)
        except nx.NetworkXNoPath:
            continue

    return new_graph


def reinject_common_join_keys_v2(
    G: ReferenceGraph,
    final: nx.DiGraph,
    nodelist: list[str],
    synonyms: set[str] = set(),
) -> bool:
    # when we've discovered a concept join, for each pair of ds nodes
    # check if they have more keys in common
    # and inject those to discovery as join conditions
    def is_ds_node(n: str) -> bool:
        return n.startswith("ds~")

    ds_graph = prune_and_merge(final, is_ds_node)
    injected = False

    for datasource in ds_graph.nodes:
        node1 = G.datasources[datasource]
        neighbors = nx.all_neighbors(ds_graph, datasource)
        for neighbor in neighbors:
            node2 = G.datasources[neighbor]
            common_concepts = set(
                x.concept.address for x in node1.columns
            ).intersection(set(x.concept.address for x in node2.columns))
            concrete_concepts = [
                x.concept for x in node1.columns if x.concept.address in common_concepts
            ]
            reduced = BuildGrain.from_concepts(concrete_concepts).components
            existing_addresses = set()
            for concrete in concrete_concepts:
                cnode = concept_to_node(concrete.with_default_grain())
                if cnode in final.nodes:
                    existing_addresses.add(concrete.address)
                    continue
            for concrete in concrete_concepts:
                if concrete.address in synonyms:
                    continue
                if concrete.address not in reduced:
                    continue
                if concrete.address in existing_addresses:
                    continue
                # skip anything that is already in the graph pseudonyms
                if any(x in concrete.pseudonyms for x in existing_addresses):
                    continue
                cnode = concept_to_node(concrete.with_default_grain())
                final.add_edge(datasource, cnode)
                final.add_edge(neighbor, cnode)
                logger.debug(
                    f"{LOGGER_PREFIX} reinjecting common join key {cnode} to list {nodelist} between {datasource} and {neighbor}, existing {existing_addresses}"
                )
                injected = True
    return injected


def determine_induced_minimal_nodes(
    G: ReferenceGraph,
    nodelist: list[str],
    environment: BuildEnvironment,
    filter_downstream: bool,
    accept_partial: bool = False,
    synonyms: set[str] = set(),
) -> nx.DiGraph | None:
    H: nx.Graph = nx.to_undirected(G).copy()

    # Add weights to edges based on target node's derivation type
    for edge in G.edges():
        _, target = edge
        target_lookup = G.concepts.get(target)

        weight = 1  # default weight
        # If either node is BASIC, set higher weight
        if target_lookup and target_lookup.derivation == Derivation.BASIC:
            if (
                isinstance(target_lookup.lineage, BuildFunction)
                and target_lookup.lineage.operator == FunctionType.ATTR_ACCESS
            ):
                weight = 1
            else:
                # raise SyntaxError(target_lookup.lineage.operator)
                weight = 50

        H.edges[edge]["weight"] = weight

    nodes_to_remove = []
    for node, lookup in G.concepts.items():
        # inclusion of aggregates can create ambiguous node relation chains
        # there may be a better way to handle this
        # can be revisited if we need to connect a derived synonym based on an aggregate
        if lookup.derivation in (
            Derivation.CONSTANT,
            Derivation.AGGREGATE,
            Derivation.FILTER,
        ):
            nodes_to_remove.append(node)
        # purge a node if we're already looking for all it's parents
        if filter_downstream and lookup.derivation not in (Derivation.ROOT,):
            nodes_to_remove.append(node)
    if nodes_to_remove:
        # logger.debug(f"Removing nodes {nodes_to_remove} from graph")
        H.remove_nodes_from(nodes_to_remove)
    isolates = list(nx.isolates(H))
    if isolates:
        # logger.debug(f"Removing isolates {isolates} from graph")
        H.remove_nodes_from(isolates)

    zero_out = list(x for x in H.nodes if G.out_degree(x) == 0 and x not in nodelist)
    while zero_out:
        logger.debug(f"Removing zero out nodes {zero_out} from graph")
        H.remove_nodes_from(zero_out)
        zero_out = list(
            x for x in H.nodes if G.out_degree(x) == 0 and x not in nodelist
        )
    try:
        # Use weight attribute for Dijkstra pathfinding
        paths = nx.multi_source_dijkstra_path(H, nodelist, weight="weight")
        # logger.debug(f"Paths found for {nodelist} {paths}")
    except nx.exception.NodeNotFound as e:
        logger.debug(f"Unable to find paths for {nodelist}- {str(e)}")
        return None
    path_removals = list(x for x in H.nodes if x not in paths)
    if path_removals:
        # logger.debug(f"Removing paths {path_removals} from graph")
        H.remove_nodes_from(path_removals)
    # logger.debug(f"Graph after path removal {H.nodes}")
    sG: nx.Graph = ax.steinertree.steiner_tree(H, nodelist, weight="weight").copy()
    if not sG.nodes:
        logger.debug(f"No Steiner tree found for nodes {nodelist}")
        return None

    logger.debug(f"Steiner tree found for nodes {nodelist} {sG.nodes}")
    final: nx.DiGraph = nx.subgraph(G, sG.nodes).copy()

    for edge in G.edges:
        if edge[1] in final.nodes and edge[0].startswith("ds~"):
            ds_name = extract_address(edge[0])
            ds = environment.datasources[ds_name]
            concept = environment.concepts[extract_address(edge[1])]
            if concept.address in [x.address for x in ds.partial_concepts]:
                if not accept_partial:
                    continue
            final.add_edge(*edge)

    reinject_common_join_keys_v2(G, final, nodelist, synonyms)

    # all concept nodes must have a parent
    if not all(
        [
            final.in_degree(node) > 0
            for node in final.nodes
            if node.startswith("c~") and node in nodelist
        ]
    ):
        missing = [
            node
            for node in final.nodes
            if node.startswith("c~") and final.in_degree(node) == 0
        ]
        logger.debug(f"Skipping graph for {nodelist} as no in_degree {missing}")
        return None

    if not all([node in final.nodes for node in nodelist]):
        missing = [node for node in nodelist if node not in final.nodes]
        logger.debug(
            f"Skipping graph for initial list {nodelist} as missing nodes {missing} from final graph {final.nodes}"
        )

        return None
    logger.debug(f"Found final graph {final.nodes}")
    return final


def canonicalize_addresses(
    reduced_concept_set: set[str], environment: BuildEnvironment
) -> set[str]:
    """
    Convert a set of concept addresses to their canonical form.
    This is necessary to ensure that we can compare concepts correctly,
    especially when dealing with aliases or pseudonyms.
    """
    return set(
        environment.concepts[x].address if x in environment.concepts else x
        for x in reduced_concept_set
    )


def detect_ambiguity_and_raise(
    all_concepts: list[BuildConcept],
    reduced_concept_sets_raw: list[set[str]],
    environment: BuildEnvironment,
) -> None:
    final_candidates: list[set[str]] = []
    common: set[str] = set()
    # find all values that show up in every join_additions
    reduced_concept_sets = [
        canonicalize_addresses(x, environment) for x in reduced_concept_sets_raw
    ]
    for ja in reduced_concept_sets:
        if not common:
            common = ja
        else:
            common = common.intersection(ja)
        if all(set(ja).issubset(y) for y in reduced_concept_sets):
            final_candidates.append(ja)
    if not final_candidates:
        filtered_paths = [x.difference(common) for x in reduced_concept_sets]
        raise AmbiguousRelationshipResolutionException(
            message=f"Multiple possible concept injections found for {[x.address for x in all_concepts]}, got {' or '.join([str(x) for x in reduced_concept_sets])}",
            parents=filtered_paths,
        )


def has_synonym(concept: BuildConcept, others: list[list[BuildConcept]]) -> bool:
    return any(
        c.address in concept.pseudonyms or concept.address in c.pseudonyms
        for sublist in others
        for c in sublist
    )


def filter_relevant_subgraphs(
    subgraphs: list[list[BuildConcept]],
) -> list[list[BuildConcept]]:
    return [
        subgraph
        for subgraph in subgraphs
        if len(subgraph) > 1
        or (
            len(subgraph) == 1
            and not has_synonym(subgraph[0], [x for x in subgraphs if x != subgraph])
        )
    ]


def filter_duplicate_subgraphs(
    subgraphs: list[list[BuildConcept]], environment
) -> list[list[BuildConcept]]:
    seen: list[set[str]] = []

    for graph in subgraphs:
        seen.append(
            canonicalize_addresses(set([x.address for x in graph]), environment)
        )
    final = []
    # sometimes w can get two subcomponents that are the same
    # due to alias resolution
    # if so, drop any that are strict subsets.
    for graph in subgraphs:
        logger.info(f"Checking graph {graph} for duplicates in {seen}")
        set_x = canonicalize_addresses(set([x.address for x in graph]), environment)
        if any([set_x.issubset(y) and set_x != y for y in seen]):
            continue
        final.append(graph)
    return final


def resolve_weak_components(
    all_concepts: List[BuildConcept],
    environment: BuildEnvironment,
    environment_graph: ReferenceGraph,
    filter_downstream: bool = True,
    accept_partial: bool = False,
    search_conditions: BuildWhereClause | None = None,
) -> list[list[BuildConcept]] | None:
    break_flag = False
    found = []
    search_graph = environment_graph.copy()
    prune_sources_for_conditions(
        search_graph, accept_partial, conditions=search_conditions
    )
    reduced_concept_sets: list[set[str]] = []

    # loop through, removing new nodes we find
    # to ensure there are not ambiguous discovery paths
    # (if we did not care about raising ambiguity errors, we could just use the first one)
    count = 0
    node_list = sorted(
        [
            concept_to_node(c.with_default_grain())
            for c in all_concepts
            if "__preql_internal" not in c.address
        ]
    )
    synonyms: set[str] = set()
    for x in all_concepts:
        synonyms = synonyms.union(x.pseudonyms)
    # from trilogy.hooks.graph_hook import GraphHook
    # GraphHook().query_graph_built(search_graph, highlight_nodes=[concept_to_node(c.with_default_grain()) for c in all_concepts if "__preql_internal" not in c.address])
    while break_flag is not True:
        count += 1
        if count > AMBIGUITY_CHECK_LIMIT:
            break_flag = True
        try:
            g = determine_induced_minimal_nodes(
                search_graph,
                node_list,
                filter_downstream=filter_downstream,
                accept_partial=accept_partial,
                environment=environment,
                synonyms=synonyms,
            )

            if not g or not g.nodes:
                break_flag = True
                continue
            if not nx.is_weakly_connected(g):
                break_flag = True
                continue
            # from trilogy.hooks.graph_hook import GraphHook
            # GraphHook().query_graph_built(g, highlight_nodes=[concept_to_node(c.with_default_grain()) for c in all_concepts if "__preql_internal" not in c.address])
            all_graph_concepts = [
                extract_concept(extract_address(node), environment)
                for node in g.nodes
                if node.startswith("c~")
            ]
            new = [x for x in all_graph_concepts if x.address not in all_concepts]

            if not new:
                break_flag = True
            # remove our new nodes for the next search path
            for n in new:
                node = concept_to_node(n)
                if node in search_graph:
                    search_graph.remove_node(node)
            # TODO: figure out better place for debugging
            # from trilogy.hooks.graph_hook import GraphHook
            # GraphHook().query_graph_built(g, highlight_nodes=[concept_to_node(c.with_default_grain()) for c in all_concepts if "__preql_internal" not in c.address])
            found.append(g)
            new_addresses = set([x.address for x in new if x.address not in synonyms])
            reduced_concept_sets.append(new_addresses)

        except nx.exception.NetworkXNoPath:
            break_flag = True
        if g and not g.nodes:
            break_flag = True
    if not found:
        return None

    detect_ambiguity_and_raise(all_concepts, reduced_concept_sets, environment)

    # take our first one as the actual graph
    g = found[0]

    subgraphs: list[list[BuildConcept]] = []
    # components = nx.strongly_connected_components(g)
    node_list = [x for x in g.nodes if x.startswith("c~")]
    components = extract_ds_components(g, node_list, environment_graph.pseudonyms)
    logger.debug(f"Extracted components {components} from {node_list}")
    for component in components:
        # we need to take unique again as different addresses may map to the same concept
        sub_component = unique(
            # sorting here is required for reproducibility
            # todo: we should sort in an optimized order
            [extract_concept(x, environment) for x in sorted(component)],
            "address",
        )
        if not sub_component:
            continue
        subgraphs.append(sub_component)
    final = filter_duplicate_subgraphs(subgraphs, environment)
    return final
    # return filter_relevant_subgraphs(subgraphs)


def subgraphs_to_merge_node(
    concept_subgraphs: list[list[BuildConcept]],
    depth: int,
    all_concepts: List[BuildConcept],
    environment,
    g,
    source_concepts,
    history,
    conditions,
    output_concepts: List[BuildConcept],
    search_conditions: BuildWhereClause | None = None,
    enable_early_exit: bool = True,
):

    parents: List[StrategyNode] = []
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} fetching subgraphs {[[c.address for c in subgraph] for subgraph in concept_subgraphs]}"
    )
    for graph in concept_subgraphs:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} fetching subgraph {[c.address for c in graph]}"
        )

        parent: StrategyNode | None = source_concepts(
            mandatory_list=graph,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
            # conditions=search_conditions,
        )
        if not parent:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} Unable to instantiate target subgraph"
            )
            return None
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} finished subgraph fetch for {[c.address for c in graph]}, have parent {type(parent)} w/ {[c.address for c in parent.output_concepts]}"
        )
        parents.append(parent)
    input_c = []
    output_c = []
    for x in parents:
        for y in x.usable_outputs:
            input_c.append(y)
            if y in output_concepts:
                output_c.append(y)
            elif any(y.address in c.pseudonyms for c in output_concepts) or any(
                c.address in y.pseudonyms for c in output_concepts
            ):
                output_c.append(y)

    if len(parents) == 1 and enable_early_exit:

        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} only one parent node, exiting early w/ {[c.address for c in parents[0].output_concepts]}"
        )
        parent = parents[0]
        return parent

    rval = MergeNode(
        input_concepts=unique(input_c, "address"),
        output_concepts=output_concepts,
        environment=environment,
        parents=parents,
        depth=depth,
        # hidden_concepts=[]
        # conditions=conditions,
        # conditions=search_conditions.conditional,
        # preexisting_conditions=search_conditions.conditional,
        # node_joins=[]
    )
    return rval


def gen_merge_node(
    all_concepts: List[BuildConcept],
    g: nx.DiGraph,
    environment: BuildEnvironment,
    depth: int,
    source_concepts,
    accept_partial: bool = False,
    history: History | None = None,
    conditions: BuildConditional | None = None,
    search_conditions: BuildWhereClause | None = None,
) -> Optional[MergeNode]:

    # we do not actually APPLY these conditions anywhere
    # though we could look at doing that as an optimization
    # it's important to include them so the base discovery loop that was generating
    # the merge node can then add them automatically
    # so we should not return a node with preexisting conditions
    if search_conditions:
        all_search_concepts = unique(
            all_concepts + list(search_conditions.row_arguments), "address"
        )
    else:
        all_search_concepts = all_concepts
    all_search_concepts = sorted(all_search_concepts, key=lambda x: x.address)
    break_set = set([x.address for x in all_search_concepts])
    for filter_downstream in [True, False]:
        weak_resolve = resolve_weak_components(
            all_search_concepts,
            environment,
            g,
            filter_downstream=filter_downstream,
            accept_partial=accept_partial,
            search_conditions=search_conditions,
        )
        if not weak_resolve:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} wasn't able to resolve graph through intermediate concept injection with accept_partial {accept_partial}, filter_downstream {filter_downstream}"
            )
            continue

        log_graph = [[y.address for y in x] for x in weak_resolve]
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Was able to resolve graph through weak component resolution - final graph {log_graph}"
        )
        for flat in log_graph:
            if set(flat) == break_set:
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} expanded concept resolution was identical to search resolution; breaking to avoid recursion error."
                )
                return None
        return subgraphs_to_merge_node(
            weak_resolve,
            depth=depth,
            all_concepts=all_search_concepts,
            environment=environment,
            g=g,
            source_concepts=source_concepts,
            history=history,
            conditions=conditions,
            search_conditions=search_conditions,
            output_concepts=all_concepts,
        )
    return None



from pytest import raises

from trilogy.constants import Parsing
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.models.environment import (
    DictImportResolver,
    Environment,
    EnvironmentOptions,
)
from trilogy.parsing.parse_engine import (
    NameShadowError,
    parse_text,
)


def test_import_shows_source():

    env = Environment(
        config=EnvironmentOptions(
            import_resolver=DictImportResolver(
                content={
                    "test": """
import test_dep as test_dep;
key x int;
datasource test (
x: x)
grain(x)
query '''
select 1 as x
union all
select 11 as x
''' TYPO
""",
                    "test_dep": """
key x int;
""",
                }
            )
        )
    )
    assert isinstance(env.config.import_resolver, DictImportResolver)

    with raises(Exception, match="Unable to import 'test', parsing error") as e:
        env.parse(
            """
        import test;
                
    select x % 10 -> x_mod_10;
                
                
    """
        )
        assert "TYPO" in str(e.value)
        assert 1 == 0


def test_concept_shadow_warning():
    x = """
key scalar int;    
property scalar.int_array list<int>;

key split <- unnest(int_array);

datasource avalues (
    int_array: int_array,
	scalar: scalar
    ) 
grain (scalar) 
query '''(
select [1,2,3,4] as int_array, 2 as scalar
union all
select [5,6,7,8] as int_array, 4 as scalar
)''';

SELECT
    int_array,
    1+2->scalar
;
"""
    with raises(NameShadowError) as e:
        env, parsed = parse_text(
            x, parse_config=Parsing(strict_name_shadow_enforcement=True)
        )
        assert "abc" in str(e)
    x = """
key scalar int;    
property scalar.int_array list<int>;

key split <- unnest(int_array);

datasource avalues (
    int_array: int_array,
	scalar: scalar
    ) 
grain (scalar) 
query '''(
select [1,2,3,4] as int_array, 2 as scalar
union all
select [5,6,7,8] as int_array, 4 as scalar
)''';

SELECT
    int_array,
    sum(scalar)->scalar
;
"""
    with raises(InvalidSyntaxException):
        env, parsed = parse_text(
            x, parse_config=Parsing(strict_name_shadow_enforcement=True)
        )


def test_parsing_bad_order():
    x = """
key scalar int;    
property scalar.int_array list<int>;

key split <- unnest(int_array);

datasource avalues (
    int_array: int_array,
	scalar: scalar
    ) 
grain (scalar) 
query '''(
select [1,2,3,4] as int_array, 2 as scalar
union all
select [5,6,7,8] as int_array, 4 as scalar
)''';

SELECT
    int_array,
    sum(scalar)->scalar
order 
    int_array asc
;
"""
    with raises(InvalidSyntaxException) as e:

        env, parsed = parse_text(
            x, parse_config=Parsing(strict_name_shadow_enforcement=True)
        )
        assert "^" in str(e)



from typing import List, Union

from pydantic import BaseModel, Field

from trilogy.core.enums import IOType
from trilogy.core.models.author import ConceptRef, HavingClause, WhereClause


class CopyQueryMixin(BaseModel):
    target: str
    target_type: IOType


class SelectTypeMixin(BaseModel):
    where_clause: Union["WhereClause", None] = Field(default=None)
    having_clause: Union["HavingClause", None] = Field(default=None)

    @property
    def output_components(self) -> List[ConceptRef]:
        raise NotImplementedError



from datetime import date, datetime
from typing import Iterable, List, Sequence, Tuple

from lark.tree import Meta

from trilogy.constants import DEFAULT_NAMESPACE, VIRTUAL_CONCEPT_PREFIX
from trilogy.core.constants import ALL_ROWS_CONCEPT
from trilogy.core.enums import (
    ConceptSource,
    Derivation,
    FunctionClass,
    FunctionType,
    Granularity,
    Modifier,
    Purpose,
    WindowType,
)
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.functions import function_args_to_output_purpose
from trilogy.core.models.author import (
    AggregateWrapper,
    AlignClause,
    AlignItem,
    CaseElse,
    CaseWhen,
    Comparison,
    Concept,
    ConceptArgs,
    ConceptRef,
    Conditional,
    FilterItem,
    Function,
    FunctionCallWrapper,
    Grain,
    HavingClause,
    ListWrapper,
    MapWrapper,
    Metadata,
    MultiSelectLineage,
    Parenthetical,
    RowsetItem,
    RowsetLineage,
    SubselectComparison,
    TraitDataType,
    TupleWrapper,
    UndefinedConcept,
    WhereClause,
    WindowItem,
    address_with_namespace,
)
from trilogy.core.models.core import DataType, arg_to_datatype
from trilogy.core.models.environment import Environment
from trilogy.core.statements.author import RowsetDerivationStatement, SelectStatement
from trilogy.utility import string_to_hash, unique

ARBITRARY_INPUTS = (
    AggregateWrapper
    | FunctionCallWrapper
    | WindowItem
    | FilterItem
    | Function
    | Parenthetical
    | ListWrapper
    | MapWrapper
    | Comparison
    | int
    | float
    | str
    | date
    | bool
)


def process_function_arg(
    arg,
    meta: Meta | None,
    environment: Environment,
):
    # if a function has an anonymous function argument
    # create an implicit concept
    if isinstance(arg, Parenthetical):
        processed = process_function_args([arg.content], meta, environment)
        return Function(
            operator=FunctionType.PARENTHETICAL,
            arguments=processed,
            output_datatype=arg_to_datatype(processed[0]),
            output_purpose=function_args_to_output_purpose(processed, environment),
        )
    elif isinstance(arg, Function):
        # if it's not an aggregate function, we can skip the virtual concepts
        # to simplify anonymous function handling
        if (
            arg.operator not in FunctionClass.AGGREGATE_FUNCTIONS.value
            and arg.operator not in FunctionClass.ONE_TO_MANY.value
        ):
            return arg
        id_hash = string_to_hash(str(arg))
        name = f"{VIRTUAL_CONCEPT_PREFIX}_{arg.operator.value}_{id_hash}"
        if f"{environment.namespace}.{name}" in environment.concepts:
            return environment.concepts[f"{environment.namespace}.{name}"]
        concept = function_to_concept(
            arg,
            name=name,
            environment=environment,
        )
        # to satisfy mypy, concept will always have metadata
        if concept.metadata and meta:
            concept.metadata.line_number = meta.line
        environment.add_concept(concept, meta=meta)
        return concept.reference
    elif isinstance(
        arg,
        (ListWrapper, MapWrapper),
    ):
        id_hash = string_to_hash(str(arg))
        name = f"{VIRTUAL_CONCEPT_PREFIX}_{id_hash}"
        if f"{environment.namespace}.{name}" in environment.concepts:
            return environment.concepts[f"{environment.namespace}.{name}"]
        concept = arbitrary_to_concept(
            arg,
            name=name,
            environment=environment,
        )
        if concept.metadata and meta:
            concept.metadata.line_number = meta.line
        environment.add_concept(concept, meta=meta)
        return concept.reference
    elif isinstance(arg, Concept):
        return arg.reference
    elif isinstance(arg, ConceptRef):
        return environment.concepts[arg.address].reference
    return arg


def process_function_args(
    args,
    meta: Meta | None,
    environment: Environment,
) -> List[ConceptRef | Function | str | int | float | date | datetime]:
    final: List[ConceptRef | Function | str | int | float | date | datetime] = []
    for arg in args:
        final.append(process_function_arg(arg, meta, environment))
    return final


def get_upstream_modifiers(
    keys: Sequence[Concept | ConceptRef], environment: Environment
) -> list[Modifier]:
    modifiers = set()
    for pkey in keys:
        if isinstance(pkey, ConceptRef):
            pkey = environment.concepts[pkey.address]
        if isinstance(pkey, UndefinedConcept):
            continue
        if pkey.modifiers:
            modifiers.update(pkey.modifiers)
    return list(modifiers)


def get_purpose_and_keys(
    purpose: Purpose | None,
    args: Tuple[ConceptRef | Concept, ...] | None,
    environment: Environment,
) -> Tuple[Purpose, set[str] | None]:
    local_purpose = purpose or function_args_to_output_purpose(args, environment)
    if local_purpose in (Purpose.PROPERTY, Purpose.METRIC) and args:
        keys = concept_list_to_keys(args, environment)
    else:
        keys = None
    return local_purpose, keys


def concept_list_to_keys(
    concepts: Tuple[Concept | ConceptRef, ...], environment: Environment
) -> set[str]:
    final_keys: List[str] = []
    for concept in concepts:

        if isinstance(concept, ConceptRef):
            concept = environment.concepts[concept.address]
        if isinstance(concept, UndefinedConcept):
            continue
        if concept.keys:
            final_keys += list(concept.keys)
        elif concept.purpose != Purpose.PROPERTY:
            final_keys.append(concept.address)
    return set(final_keys)


def constant_to_concept(
    parent: (
        ListWrapper | TupleWrapper | MapWrapper | int | float | str | date | datetime
    ),
    name: str,
    namespace: str,
    metadata: Metadata | None = None,
) -> Concept:
    const_function: Function = Function(
        operator=FunctionType.CONSTANT,
        output_datatype=arg_to_datatype(parent),
        output_purpose=Purpose.CONSTANT,
        arguments=[parent],
    )
    # assert const_function.arguments[0] == parent, f'{const_function.arguments[0]} != {parent}, {type(const_function.arguments[0])} != {type(parent)}'
    fmetadata = metadata or Metadata()
    return Concept(
        name=name,
        datatype=const_function.output_datatype,
        purpose=Purpose.CONSTANT,
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
        lineage=const_function,
        grain=Grain(),
        namespace=namespace,
        metadata=fmetadata,
    )


def atom_is_relevant(
    atom,
    others: list[Concept | ConceptRef],
    environment: Environment | None = None,
):

    if isinstance(atom, (ConceptRef, Concept)):
        # when we are looking at atoms, if there is a concept that is in others
        # return directly
        if atom.address in others:
            return False
        return concept_is_relevant(atom, others, environment)

    if isinstance(atom, AggregateWrapper) and not atom.by:
        return False
    elif isinstance(atom, AggregateWrapper):
        return any(atom_is_relevant(x, others, environment) for x in atom.by)

    elif isinstance(atom, Function):
        relevant = False
        for arg in atom.arguments:

            relevant = relevant or atom_is_relevant(arg, others, environment)
        return relevant
    elif isinstance(atom, FunctionCallWrapper):
        return any(
            [atom_is_relevant(atom.content, others, environment)]
            + [atom_is_relevant(x, others, environment) for x in atom.args]
        )
    elif isinstance(atom, CaseWhen):
        rval = atom_is_relevant(atom.expr, others, environment) or atom_is_relevant(
            atom.comparison, others, environment
        )
        return rval
    elif isinstance(atom, CaseElse):

        rval = atom_is_relevant(atom.expr, others, environment)
        return rval
    elif isinstance(atom, SubselectComparison):
        return atom_is_relevant(atom.left, others, environment)
    elif isinstance(atom, Comparison):
        return atom_is_relevant(atom.left, others, environment) or atom_is_relevant(
            atom.right, others, environment
        )
    elif isinstance(atom, Conditional):
        return atom_is_relevant(atom.left, others, environment) or atom_is_relevant(
            atom.right, others, environment
        )
    elif isinstance(atom, Parenthetical):
        return atom_is_relevant(atom.content, others, environment)
    elif isinstance(atom, ConceptArgs):
        # use atom is relevant here to trigger the early exit behavior for concepts in set
        return any(
            [atom_is_relevant(x, others, environment) for x in atom.concept_arguments]
        )

    return False


def concept_is_relevant(
    concept: Concept | ConceptRef,
    others: list[Concept | ConceptRef],
    environment: Environment | None = None,
) -> bool:
    if isinstance(concept, UndefinedConcept):
        return False
    if concept.datatype == DataType.UNKNOWN:
        return False

    if isinstance(concept, ConceptRef):
        if environment:
            concept = environment.concepts[concept.address]
        else:
            raise SyntaxError(
                "Require environment to determine relevance of ConceptRef"
            )
    if concept.derivation == Derivation.CONSTANT:
        return False
    if concept.is_aggregate and not (
        isinstance(concept.lineage, AggregateWrapper) and concept.lineage.by
    ):

        return False
    if concept.purpose in (Purpose.PROPERTY, Purpose.METRIC) and concept.keys:
        if all([c in others for c in concept.keys]):
            return False
    if (
        concept.purpose == Purpose.KEY
        and concept.keys
        and all([c in others for c in concept.keys])
    ):
        return False
    if concept.purpose in (Purpose.METRIC,):
        if all([c in others for c in concept.grain.components]):
            return False
    if (
        concept.derivation in (Derivation.BASIC,)
        and isinstance(concept.lineage, Function)
        and concept.lineage.operator == FunctionType.DATE_SPINE
    ):
        return True
    if concept.derivation in (Derivation.BASIC,) and isinstance(
        concept.lineage, (Function, CaseWhen)
    ):
        relevant = False
        for arg in concept.lineage.arguments:
            relevant = atom_is_relevant(arg, others, environment) or relevant
        return relevant
    if concept.derivation in (Derivation.BASIC,) and isinstance(
        concept.lineage, Parenthetical
    ):
        return atom_is_relevant(concept.lineage.content, others, environment)

    if concept.granularity == Granularity.SINGLE_ROW:
        return False
    return True


def concepts_to_grain_concepts(
    concepts: Iterable[Concept | ConceptRef | str],
    environment: Environment | None,
    local_concepts: dict[str, Concept] | None = None,
) -> set[str]:
    preconcepts: list[Concept] = []
    for c in concepts:
        if isinstance(c, Concept):
            preconcepts.append(c)

        elif isinstance(c, ConceptRef) and environment:
            if local_concepts and c.address in local_concepts:
                preconcepts.append(local_concepts[c.address])
            else:
                preconcepts.append(environment.concepts[c.address])
        elif isinstance(c, str) and environment:
            if local_concepts and c in local_concepts:
                preconcepts.append(local_concepts[c])
            else:
                preconcepts.append(environment.concepts[c])
        else:
            raise ValueError(
                f"Unable to resolve input {c} without environment provided to concepts_to_grain call"
            )
    pconcepts: list[Concept] = []
    for x in preconcepts:
        if (
            x.lineage
            and isinstance(x.lineage, Function)
            and x.lineage.operator == FunctionType.ALIAS
        ):
            # if the function is an alias, use the unaliased concept to calculate grain
            pconcepts.append(environment.concepts[x.lineage.arguments[0].address])  # type: ignore
        else:
            pconcepts.append(x)

    seen = set()
    for sub in pconcepts:
        if sub.address in seen:
            continue
        if not concept_is_relevant(sub, pconcepts, environment):  # type: ignore

            continue
        seen.add(sub.address)

    return seen


def _get_relevant_parent_concepts(arg) -> tuple[list[ConceptRef], bool]:
    from trilogy.core.models.author import get_concept_arguments

    is_metric = False
    if isinstance(arg, Function):
        all = []
        for y in arg.arguments:
            refs, local_flag = get_relevant_parent_concepts(y)
            all += refs
            is_metric = is_metric or local_flag
        return all, is_metric
    elif isinstance(arg, AggregateWrapper) and not arg.by:
        return [], True
    elif isinstance(arg, AggregateWrapper) and arg.by:
        return [x.reference for x in arg.by], True
    elif isinstance(arg, FunctionCallWrapper):
        return get_relevant_parent_concepts(arg.content)
    elif isinstance(arg, Comparison):
        left, lflag = get_relevant_parent_concepts(arg.left)
        right, rflag = get_relevant_parent_concepts(arg.right)
        return left + right, lflag or rflag
    return get_concept_arguments(arg), False


def get_relevant_parent_concepts(arg) -> tuple[list[ConceptRef], bool]:
    concepts, status = _get_relevant_parent_concepts(arg)
    return unique(concepts, "address"), status


def group_function_to_concept(
    parent: Function,
    name: str,
    environment: Environment,
    namespace: str | None = None,
    metadata: Metadata | None = None,
):
    pkeys: List[Concept] = []
    namespace = namespace or environment.namespace
    is_metric = False
    ref_args, is_metric = get_relevant_parent_concepts(parent)
    concrete_args = [environment.concepts[c.address] for c in ref_args]
    pkeys += [x for x in concrete_args if not x.derivation == Derivation.CONSTANT]
    modifiers = get_upstream_modifiers(pkeys, environment)
    key_grain: list[str] = []
    for x in pkeys:
        # for a group to, if we have a dynamic metric, ignore it
        # it will end up with the group target grain
        if x.purpose == Purpose.METRIC and not x.keys:
            continue
        # metrics will group to keys, so do no do key traversal
        elif is_metric:
            key_grain.append(x.address)
        else:
            key_grain.append(x.address)
    keys = set(key_grain)

    grain = Grain.from_concepts(keys, environment)
    if is_metric:
        purpose = Purpose.METRIC
    elif not pkeys:
        purpose = Purpose.CONSTANT
    else:
        purpose = parent.output_purpose
    fmetadata = metadata or Metadata()
    granularity = Granularity.MULTI_ROW

    if grain is not None:
        # deduplicte
        grain = Grain.from_concepts(grain.components, environment)

        r = Concept(
            name=name,
            datatype=parent.output_datatype,
            purpose=purpose,
            lineage=parent,
            namespace=namespace,
            keys=keys,
            modifiers=modifiers,
            grain=grain,
            metadata=fmetadata,
            derivation=Derivation.GROUP_TO,
            granularity=granularity,
        )
        return r

    return Concept(
        name=name,
        datatype=parent.output_datatype,
        purpose=purpose,
        lineage=parent,
        namespace=namespace,
        keys=keys,
        modifiers=modifiers,
        metadata=fmetadata,
        derivation=Derivation.BASIC,
        granularity=granularity,
    )


def function_to_concept(
    parent: Function,
    name: str,
    environment: Environment,
    namespace: str | None = None,
    metadata: Metadata | None = None,
) -> Concept:

    pkeys: List[Concept] = []
    namespace = namespace or environment.namespace
    is_metric = False
    ref_args, is_metric = get_relevant_parent_concepts(parent)
    concrete_args = [environment.concepts[c.address] for c in ref_args]
    pkeys += [
        x
        for x in concrete_args
        if not x.derivation == Derivation.CONSTANT
        and not (x.derivation == Derivation.AGGREGATE and not x.grain.components)
    ]
    grain: Grain | None = Grain()
    for x in pkeys:
        grain += x.grain
    if parent.operator in FunctionClass.ONE_TO_MANY.value:
        # if the function will create more rows, we don't know what grain this is at
        grain = None
    modifiers = get_upstream_modifiers(pkeys, environment)
    key_grain: list[str] = []
    for x in pkeys:
        # metrics will group to keys, so do not do key traversal
        if is_metric:
            key_grain.append(x.address)
        # otherwse, for row ops, assume keys are transitive
        elif x.keys:
            key_grain += [*x.keys]
        else:
            key_grain.append(x.address)
    keys = set(key_grain)
    if is_metric:
        purpose = Purpose.METRIC
    elif not pkeys:
        purpose = Purpose.CONSTANT
    else:
        purpose = parent.output_purpose
    fmetadata = metadata or Metadata()
    if parent.operator in FunctionClass.AGGREGATE_FUNCTIONS.value:
        derivation = Derivation.AGGREGATE
        if (
            grain
            and grain.components
            and all(x.endswith(ALL_ROWS_CONCEPT) for x in grain.components)
        ):
            granularity = Granularity.SINGLE_ROW
        else:
            granularity = Granularity.MULTI_ROW
    elif parent.operator == FunctionType.UNION:
        derivation = Derivation.UNION
        granularity = Granularity.MULTI_ROW
    elif parent.operator in FunctionClass.ONE_TO_MANY.value:
        derivation = Derivation.UNNEST
        granularity = Granularity.MULTI_ROW
    elif parent.operator == FunctionType.RECURSE_EDGE:
        derivation = Derivation.RECURSIVE
        granularity = Granularity.MULTI_ROW
    elif parent.operator in FunctionClass.SINGLE_ROW.value:
        derivation = Derivation.CONSTANT
        granularity = Granularity.SINGLE_ROW
    elif concrete_args and all(
        x.derivation == Derivation.CONSTANT for x in concrete_args
    ):
        derivation = Derivation.CONSTANT
        granularity = Granularity.SINGLE_ROW
    else:
        derivation = Derivation.BASIC
        granularity = Granularity.MULTI_ROW
    if grain is not None:
        r = Concept(
            name=name,
            datatype=parent.output_datatype,
            purpose=purpose,
            lineage=parent,
            namespace=namespace,
            keys=keys,
            modifiers=modifiers,
            grain=grain,
            metadata=fmetadata,
            derivation=derivation,
            granularity=granularity,
        )
        return r

    return Concept(
        name=name,
        datatype=parent.output_datatype,
        purpose=purpose,
        lineage=parent,
        namespace=namespace,
        keys=keys,
        modifiers=modifiers,
        metadata=fmetadata,
        derivation=derivation,
        granularity=granularity,
    )


def filter_item_to_concept(
    parent: FilterItem,
    name: str,
    namespace: str,
    environment: Environment,
    metadata: Metadata | None = None,
) -> Concept:
    fmetadata = metadata or Metadata()
    fallback_keys = set()
    if isinstance(parent.content, ConceptRef):
        cparent = environment.concepts[parent.content.address]
        fallback_keys = set([cparent.address])
    elif isinstance(
        parent.content,
        (
            FilterItem,
            AggregateWrapper,
            FunctionCallWrapper,
            WindowItem,
            Function,
            ListWrapper,
            MapWrapper,
            int,
            str,
            float,
        ),
    ):
        cparent = arbitrary_to_concept(parent.content, environment, namespace=namespace)

    else:
        raise NotImplementedError(
            f"Filter item with non ref content {parent.content} ({type(parent.content)}) not yet supported"
        )
    modifiers = get_upstream_modifiers(
        cparent.concept_arguments, environment=environment
    )
    grain = cparent.grain if cparent.purpose == Purpose.PROPERTY else Grain()
    granularity = cparent.granularity
    return Concept(
        name=name,
        datatype=cparent.datatype,
        purpose=Purpose.PROPERTY,
        lineage=parent,
        metadata=fmetadata,
        namespace=namespace,
        # filtered copies cannot inherit keys
        keys=(cparent.keys if cparent.purpose == Purpose.PROPERTY else fallback_keys),
        grain=grain,
        modifiers=modifiers,
        derivation=Derivation.FILTER,
        granularity=granularity,
    )


def window_item_to_concept(
    parent: WindowItem,
    name: str,
    namespace: str,
    environment: Environment,
    metadata: Metadata | None = None,
) -> Concept:
    fmetadata = metadata or Metadata()
    if not isinstance(parent.content, ConceptRef):
        raise NotImplementedError(
            f"Window function with non ref content {parent.content} not yet supported"
        )
    bcontent = environment.concepts[parent.content.address]
    if isinstance(bcontent, UndefinedConcept):
        return UndefinedConcept(address=f"{namespace}.{name}", metadata=fmetadata)
    if bcontent.purpose == Purpose.METRIC:
        local_purpose, keys = get_purpose_and_keys(None, (bcontent,), environment)
    else:
        local_purpose = Purpose.PROPERTY
        keys = Grain.from_concepts(
            [bcontent.address] + [y.address for y in parent.over], environment
        ).components

    # when including the order by in discovery grain
    if parent.order_by:

        grain_components = parent.over + [bcontent.output]
        for item in parent.order_by:
            relevant, _ = get_relevant_parent_concepts(item.expr)
            grain_components += relevant
    else:
        grain_components = parent.over + [bcontent.output]

    final_grain = Grain.from_concepts(grain_components, environment)
    modifiers = get_upstream_modifiers(bcontent.concept_arguments, environment)
    datatype = parent.content.datatype
    if parent.type in (
        # WindowType.RANK,
        WindowType.ROW_NUMBER,
        WindowType.COUNT,
        WindowType.COUNT_DISTINCT,
    ):
        datatype = DataType.INTEGER
    if parent.type == WindowType.RANK:
        datatype = TraitDataType(type=DataType.INTEGER, traits=["rank"])
    return Concept(
        name=name,
        datatype=datatype,
        purpose=local_purpose,
        lineage=parent,
        metadata=fmetadata,
        grain=final_grain,
        namespace=namespace,
        keys=keys,
        modifiers=modifiers,
        derivation=Derivation.WINDOW,
        granularity=bcontent.granularity,
    )


def agg_wrapper_to_concept(
    parent: AggregateWrapper,
    namespace: str,
    name: str,
    environment: Environment,
    metadata: Metadata | None = None,
) -> Concept:
    _, keys = get_purpose_and_keys(
        Purpose.METRIC, tuple(parent.by) if parent.by else None, environment=environment
    )
    # anything grouped to a grain should be a property
    # at that grain
    fmetadata = metadata or Metadata()
    aggfunction = parent.function
    modifiers = get_upstream_modifiers(parent.concept_arguments, environment)
    grain = Grain.from_concepts(parent.by, environment) if parent.by else Grain()
    granularity = Concept.calculate_granularity(Derivation.AGGREGATE, grain, parent)

    out = Concept(
        name=name,
        datatype=aggfunction.output_datatype,
        purpose=Purpose.METRIC,
        metadata=fmetadata,
        lineage=parent,
        grain=grain,
        namespace=namespace,
        keys=set([x.address for x in parent.by]) if parent.by else keys,
        modifiers=modifiers,
        derivation=Derivation.AGGREGATE,
        granularity=granularity,
    )
    for x in parent.function.concept_arguments:
        if x.address == out.address:
            raise InvalidSyntaxException(
                f"Aggregate concept {out.address} cannot reference itself. If defining a new concept in a select, use a new name."
            )
    return out


def align_item_to_concept(
    parent: AlignItem,
    align_clause: AlignClause,
    selects: list[SelectStatement],
    environment: Environment,
    where: WhereClause | None = None,
    having: HavingClause | None = None,
    limit: int | None = None,
) -> Concept:
    align = parent
    datatypes = set([c.datatype for c in align.concepts])
    if len(datatypes) > 1:
        raise InvalidSyntaxException(
            f"Datatypes do not align for merged statements {align.alias}, have {datatypes}"
        )

    new_selects = [x.as_lineage(environment) for x in selects]
    multi_lineage = MultiSelectLineage(
        selects=new_selects,
        align=align_clause,
        namespace=align.namespace,
        where_clause=where,
        having_clause=having,
        limit=limit,
        hidden_components=set(y for x in new_selects for y in x.hidden_components),
    )
    grain = Grain()
    new = Concept(
        name=align.alias,
        datatype=datatypes.pop(),
        purpose=Purpose.PROPERTY,
        lineage=multi_lineage,
        grain=grain,
        namespace=align.namespace,
        granularity=Granularity.MULTI_ROW,
        derivation=Derivation.MULTISELECT,
        keys=set(x.address for x in align.concepts),
    )
    return new


def derive_item_to_concept(
    parent: ARBITRARY_INPUTS,
    name: str,
    lineage: MultiSelectLineage,
    namespace: str | None = None,
) -> Concept:
    datatype = arg_to_datatype(parent)
    grain = Grain()
    new = Concept(
        name=name,
        datatype=datatype,
        purpose=Purpose.PROPERTY,
        lineage=lineage,
        grain=grain,
        namespace=namespace or DEFAULT_NAMESPACE,
        granularity=Granularity.MULTI_ROW,
        derivation=Derivation.MULTISELECT,
    )
    return new


def rowset_concept(
    orig_address: ConceptRef,
    environment: Environment,
    rowset: RowsetDerivationStatement,
    pre_output: list[Concept],
    orig: dict[str, Concept],
    orig_map: dict[str, Concept],
):
    orig_concept = environment.concepts[orig_address.address]
    name = orig_concept.name
    if isinstance(orig_concept.lineage, FilterItem):
        if orig_concept.lineage.where == rowset.select.where_clause and isinstance(
            orig_concept.lineage.content, (ConceptRef, Concept)
        ):
            name = environment.concepts[orig_concept.lineage.content.address].name
    base_namespace = (
        f"{rowset.name}.{orig_concept.namespace}"
        if orig_concept.namespace != rowset.namespace
        else rowset.name
    )

    new_concept = Concept(
        name=name,
        datatype=orig_concept.datatype,
        purpose=orig_concept.purpose,
        lineage=None,
        grain=orig_concept.grain,
        metadata=Metadata(concept_source=ConceptSource.CTE),
        namespace=base_namespace,
        keys=orig_concept.keys,
        derivation=Derivation.ROWSET,
        granularity=orig_concept.granularity,
        pseudonyms={
            address_with_namespace(x, rowset.name) for x in orig_concept.pseudonyms
        },
    )
    for x in orig_concept.pseudonyms:
        new_address = address_with_namespace(x, rowset.name)
        origa = environment.alias_origin_lookup[x]
        environment.concepts[new_address] = new_concept
        environment.alias_origin_lookup[new_address] = origa.model_copy(
            update={"namespace": f"{rowset.name}.{origa.namespace}"}
        )
    orig[orig_concept.address] = new_concept
    orig_map[new_concept.address] = orig_concept
    pre_output.append(new_concept)


def rowset_to_concepts(rowset: RowsetDerivationStatement, environment: Environment):
    pre_output: list[Concept] = []
    orig: dict[str, Concept] = {}
    orig_map: dict[str, Concept] = {}
    for orig_address in rowset.select.output_components:
        rowset_concept(orig_address, environment, rowset, pre_output, orig, orig_map)
    select_lineage = rowset.select.as_lineage(environment)
    for x in pre_output:
        x.lineage = RowsetItem(
            content=orig_map[x.address].reference,
            rowset=RowsetLineage(
                name=rowset.name,
                derived_concepts=[x.reference for x in pre_output],
                select=select_lineage,
            ),
        )
    default_grain = Grain.from_concepts([*pre_output])
    # remap everything to the properties of the rowset
    for x in pre_output:
        if x.keys:
            if all([k in orig for k in x.keys]):
                x.keys = set([orig[k].address if k in orig else k for k in x.keys])
            else:
                # TODO: fix this up
                x.keys = set()
        if all([c in orig for c in x.grain.components]):
            x.grain = Grain(components={orig[c].address for c in x.grain.components})
        else:
            x.grain = default_grain
    return pre_output


def generate_concept_name(
    parent: ARBITRARY_INPUTS,
) -> str:
    """Generate a name for a concept based on its parent type and content."""
    if isinstance(parent, AggregateWrapper):
        return f"{VIRTUAL_CONCEPT_PREFIX}_agg_{parent.function.operator.value}_{string_to_hash(str(parent))}"
    elif isinstance(parent, WindowItem):
        return f"{VIRTUAL_CONCEPT_PREFIX}_window_{parent.type.value}_{string_to_hash(str(parent))}"
    elif isinstance(parent, FilterItem):
        if isinstance(parent.content, ConceptRef):
            return f"{VIRTUAL_CONCEPT_PREFIX}_filter_{parent.content.name}_{string_to_hash(str(parent))}"
        else:
            return f"{VIRTUAL_CONCEPT_PREFIX}_filter_{string_to_hash(str(parent))}"
    elif isinstance(parent, Function):
        if parent.operator == FunctionType.GROUP:
            return f"{VIRTUAL_CONCEPT_PREFIX}_group_to_{string_to_hash(str(parent))}"
        else:
            return f"{VIRTUAL_CONCEPT_PREFIX}_func_{parent.operator.value}_{string_to_hash(str(parent))}"
    elif isinstance(parent, Parenthetical):
        return f"{VIRTUAL_CONCEPT_PREFIX}_paren_{string_to_hash(str(parent))}"
    elif isinstance(parent, FunctionCallWrapper):
        return f"{VIRTUAL_CONCEPT_PREFIX}_{parent.name}_{string_to_hash(str(parent))}"
    elif isinstance(parent, Comparison):
        return f"{VIRTUAL_CONCEPT_PREFIX}_comp_{string_to_hash(str(parent))}"
    else:  # ListWrapper, MapWrapper, or primitive types
        return f"{VIRTUAL_CONCEPT_PREFIX}_{string_to_hash(str(parent))}"


def parenthetical_to_concept(
    parent: Parenthetical,
    name: str,
    namespace: str,
    environment: Environment,
    metadata: Metadata | None = None,
) -> Concept:
    if isinstance(
        parent.content,
        ARBITRARY_INPUTS,
    ):

        return arbitrary_to_concept(
            parent.content, environment, namespace, name, metadata
        )
    raise NotImplementedError(
        f"Parenthetical with non-supported content {parent.content} ({type(parent.content)}) not yet supported"
    )


def comparison_to_concept(
    parent: Comparison,
    name: str,
    namespace: str,
    environment: Environment,
    metadata: Metadata | None = None,
):
    fmetadata = metadata or Metadata()

    pkeys: List[Concept] = []
    namespace = namespace or environment.namespace
    is_metric = False
    ref_args, is_metric = get_relevant_parent_concepts(parent)
    concrete_args = [environment.concepts[c.address] for c in ref_args]
    pkeys += [
        x
        for x in concrete_args
        if not x.derivation == Derivation.CONSTANT
        and not (x.derivation == Derivation.AGGREGATE and not x.grain.components)
    ]
    grain: Grain | None = Grain()
    for x in pkeys:
        grain += x.grain
    if parent.operator in FunctionClass.ONE_TO_MANY.value:
        # if the function will create more rows, we don't know what grain this is at
        grain = None
    modifiers = get_upstream_modifiers(pkeys, environment)
    key_grain: list[str] = []
    for x in pkeys:
        # metrics will group to keys, so do not do key traversal
        if is_metric:
            key_grain.append(x.address)
        # otherwse, for row ops, assume keys are transitive
        elif x.keys:
            key_grain += [*x.keys]
        else:
            key_grain.append(x.address)
    keys = set(key_grain)
    if is_metric:
        purpose = Purpose.METRIC
    elif not pkeys:
        purpose = Purpose.CONSTANT
    else:
        purpose = Purpose.PROPERTY
    fmetadata = metadata or Metadata()

    if grain is not None:
        r = Concept(
            name=name,
            datatype=parent.output_datatype,
            purpose=purpose,
            lineage=parent,
            namespace=namespace,
            keys=keys,
            modifiers=modifiers,
            grain=grain,
            metadata=fmetadata,
            derivation=Derivation.BASIC,
            granularity=Granularity.MULTI_ROW,
        )
        return r

    return Concept(
        name=name,
        datatype=parent.output_datatype,
        purpose=purpose,
        lineage=parent,
        namespace=namespace,
        keys=keys,
        modifiers=modifiers,
        metadata=fmetadata,
        derivation=Derivation.BASIC,
        granularity=Granularity.MULTI_ROW,
    )


def arbitrary_to_concept(
    parent: ARBITRARY_INPUTS,
    environment: Environment,
    namespace: str | None = None,
    name: str | None = None,
    metadata: Metadata | None = None,
) -> Concept:
    namespace = namespace or environment.namespace

    # this is purely for the parse tree, discard from derivation
    if isinstance(parent, FunctionCallWrapper):
        return arbitrary_to_concept(
            parent.content, environment, namespace, name, metadata  # type: ignore
        )

    # Generate name if not provided
    if not name:
        name = generate_concept_name(parent)

    if isinstance(parent, AggregateWrapper):
        return agg_wrapper_to_concept(
            parent, namespace, name, metadata=metadata, environment=environment
        )
    elif isinstance(parent, WindowItem):
        return window_item_to_concept(
            parent,
            name,
            namespace,
            environment=environment,
            metadata=metadata,
        )
    elif isinstance(parent, FilterItem):
        return filter_item_to_concept(
            parent,
            name,
            namespace,
            environment=environment,
            metadata=metadata,
        )
    elif isinstance(parent, Function):
        if parent.operator == FunctionType.GROUP:
            return group_function_to_concept(
                parent,
                name,
                environment=environment,
                namespace=namespace,
                metadata=metadata,
            )
        return function_to_concept(
            parent,
            name,
            metadata=metadata,
            environment=environment,
            namespace=namespace,
        )
    elif isinstance(parent, ListWrapper):
        return constant_to_concept(parent, name, namespace, metadata)
    elif isinstance(parent, Parenthetical):
        return parenthetical_to_concept(parent, name, namespace, environment, metadata)
    elif isinstance(parent, Comparison):
        return comparison_to_concept(parent, name, namespace, environment, metadata)
    else:
        return constant_to_concept(parent, name, namespace, metadata)



from pathlib import Path

from pytest import raises

from trilogy import Dialects
from trilogy.core.enums import Modifier, Purpose
from trilogy.core.exceptions import (
    FrozenEnvironmentException,
    UndefinedConceptException,
)
from trilogy.core.models.author import Concept
from trilogy.core.models.core import DataType
from trilogy.core.models.environment import Environment, LazyEnvironment


def test_concept_rehydration():
    test = {
        "name": "order_timestamp.month_start",
        "datatype": "date",
        "purpose": "property",
        "derivation": "root",
        "granularity": "multi_row",
        "metadata": {
            "description": "Auto-derived from a local.order_timestamp. The date truncated to the month.",
            "line_number": None,
            "concept_source": "auto_derived",
        },
        "lineage": {
            "operator": "date_truncate",
            "arg_count": 2,
            "output_datatype": "date",
            "output_purpose": "property",
            "valid_inputs": [
                ["datetime", "string", "timestamp", "date"],
                ["date_part"],
            ],
            "arguments": [
                {
                    "address": "local.order_timestamp",
                    "datatype": "timestamp",
                    "metadata": {
                        "description": None,
                        "line_number": None,
                        "concept_source": "manual",
                    },
                },
                "month",
            ],
        },
        "namespace": "local",
        "keys": ["local.order_timestamp"],
        "grain": {"components": ["local.order_id"], "where_clause": None},
        "modifiers": [],
        "pseudonyms": [],
    }
    Concept.model_validate(test)
    test["lineage"]["arguments"] = [
        {
            "address": "local.order_timestamp",
            "datatype": "timestamp",
            "metadata": {
                "description": None,
                "line_number": None,
                "concept_source": "manual",
            },
        },
        "monthz",
    ]
    with raises(TypeError):
        Concept.model_validate(test)


def test_environment_serialization(test_environment: Environment):

    path = test_environment.to_cache()
    test_environment2 = Environment.from_cache(path)
    assert test_environment2

    assert test_environment.concepts == test_environment2.concepts
    assert test_environment.datasources == test_environment2.datasources
    for k, v in test_environment.concepts.items():
        assert v == test_environment2.concepts[k]


def test_environment_from_path():
    env = Environment.from_file(Path(__file__).parent / "test_env.preql")

    assert "local.id" in env.concepts


def test_lazy_environment_from_path():
    env = LazyEnvironment(load_path=Path(__file__).parent / "test_env.preql")

    assert not env.loaded

    _ = len(env.concepts)

    assert env.loaded

    env2 = Environment.from_file(Path(__file__).parent / "test_env.preql")

    assert env.concepts == env2.concepts


def test_frozen_environment():
    env = Environment.from_file(Path(__file__).parent / "test_env.preql")

    exec = Dialects.DUCK_DB.default_executor(environment=env)

    run = exec.execute_query(
        """select id+5 ->id_plus_5 order by id_plus_5 asc;"""
    ).fetchall()
    assert run[0].id_plus_5 == 6
    env.freeze()
    with raises(FrozenEnvironmentException):
        env.add_concept(
            Concept(name="test", datatype=DataType.INTEGER, purpose=Purpose.KEY)
        )


def test_environment_invalid():
    env = Environment()
    env.concepts.fail_on_missing = False
    x = env.concepts["abc"]
    assert x.name == "abc"

    env.concepts.fail_on_missing = True
    try:
        x = env.concepts["abc"]
        assert 1 == 0
    except Exception as e:
        assert isinstance(e, UndefinedConceptException)


def test_environment_merge():
    env1: Environment
    env1, _ = Environment().parse(
        """
key  order_id int;   
                                
                                datasource orders
                                (order_id:order_id)
                                grain (order_id)
                                address orders;
"""
    )

    env2, _ = Environment().parse(
        """
                                key order_id int;
                                                                
                                datasource replacements
                                (order_id:order_id)
                                grain (order_id)
                                address replacements;

                                
"""
    )

    env1.add_import("replacements", env2)

    _ = env1.merge_concept(
        env1.concepts["replacements.order_id"],
        env1.concepts["order_id"],
        modifiers=[Modifier.PARTIAL],
    )

    assert env1.concepts["order_id"] == env1.concepts["replacements.order_id"]

    found = False
    for x in env1.datasources["replacements.replacements"].columns:
        if (
            x.alias == "order_id"
            and x.concept.address == env1.concepts["order_id"].address
        ):
            assert x.concept == env1.concepts["order_id"]
            assert x.modifiers == [Modifier.PARTIAL]
            found = True
    assert found


def test_environment_select_promotion():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
const x <- 6;

select x+2 as y;

select y;            
                    """
    ).fetchall()

    assert results[0].y == 8
